import numpy as np
import scipy.stats

from . import utils as cp_utils
from . import lnorm_libs


def ppf(x, p=np.arange(0.1, 1.0, 0.1), means=False, waicscores=False, logscores=False, rust=False, nrust=100000, debug=False):
    """
    Log-normal Distribution Predictions Based on a Calibrating Prior
    
    Parameters
    ----------
    x : array-like
        Training data values
    p : array-like, optional
        Probabilities for quantiles (default: np.arange(0.1, 1.0, 0.1))
    means : bool
        Whether to calculate means (default False)
    waicscores : bool
        Whether to calculate WAIC scores (default False)
    logscores : bool
        Whether to calculate log scores (default False)
    rust : bool
        Whether to use RUST method (default False)
    nrust : int
        Number of RUST samples (default 100000)
    debug : bool
        Debug flag (default False)
        
    Returns
    -------
    dict
        Dictionary containing quantiles and related statistics
    """
    
    # Input validation
    x = cp_utils.to_array(x)
    p = cp_utils.to_array(p)
    assert np.all(np.isfinite(x)) and np.all(~np.isnan(x)), "x must be finite and not NaN"
    assert np.all(np.isfinite(p)) and np.all(~np.isnan(p)), "p must be finite and not NaN"
    assert np.all(p > 0) and np.all(p < 1), "p must be between 0 and 1"
    assert np.all(x >= 0), "x must be non-negative"
    
    alpha = 1 - p
    y = np.log(x)
    nx = len(x)
    nalpha = len(alpha)
    
    # 2 ml param estimate
    ml_params = lnorm_libs.norm_ml_params(y)  # note that it uses y, and the normal routine
    v1hat = ml_params[0]
    v2hat = ml_params[1]
    
    # 3 aic
    ml_value = np.sum(scipy.stats.lognorm.logpdf(x, s=v2hat, scale=np.exp(v1hat)))
    maic = cp_utils.make_maic(ml_value, nparams=2)
    
    # 4 ml quantiles (vectorized over alpha)
    ml_quantiles = scipy.stats.lognorm.ppf(1-alpha, s=v2hat, scale=np.exp(v1hat))
    
    ldd = "only relevant for DMGS models, not analytic models"
    lddi = "only relevant for DMGS models, not analytic models"
    expinfmat = np.zeros((2, 2))
    expinfmat[0, 0] = nx / (v2hat * v2hat)
    expinfmat[1, 1] = 2 * nx / (v2hat * v2hat)
    expinfmati = np.linalg.inv(expinfmat)
    standard_errors = np.zeros(2)
    standard_errors[0] = np.sqrt(expinfmati[0, 0])
    standard_errors[1] = np.sqrt(expinfmati[1, 1])
    
    # 5 rhp quantiles
    mu = np.mean(y)
    # calculate the unbiased variance
    s1 = np.sqrt(np.var(y, ddof=1))  # ddof=1 for unbiased variance like R's var()
    temp = scipy.stats.t.ppf(1-alpha, df=nx-1)
    # convert the unbiased to predictive
    rh_quantiles = np.exp(mu + temp * s1 * np.sqrt(1 + 1/nx))
    
    # 6 means (might as well always calculate)
    ml_mean = np.exp(v1hat + 0.5 * v2hat * v2hat)
    rh_mean = "no analytic expression"
    
    # 7 waicscores
    waic = lnorm_libs.lnorm_waic(waicscores, x, v1hat, v2hat)
    waic1 = waic['waic1']
    waic2 = waic['waic2']
    
    # 8 logscores
    logscores_result = lnorm_libs.lnorm_logscores(logscores, x)
    ml_oos_logscore = logscores_result['ml_oos_logscore']
    rh_oos_logscore = logscores_result['rh_oos_logscore']
    
    # 9 rust
    ru_quantiles = "rust not selected"
    if rust:
        rustsim = rvs(nrust, x, rust=True, mlcp=False)
        ru_quantiles = cp_utils.makeq(rustsim['ru_deviates'], p)
    
    # return
    return {
        'ml_params': ml_params,
        'ml_value': ml_value,
        'standard_errors': standard_errors,
        'ml_quantiles': ml_quantiles,
        'cp_quantiles': rh_quantiles,
        'ru_quantiles': ru_quantiles,
        'maic': maic,
        'waic1': waic1,
        'waic2': waic2,
        'ml_oos_logscore': ml_oos_logscore,
        'cp_oos_logscore': rh_oos_logscore,
        'ml_mean': ml_mean,
        'cp_mean': rh_mean,
        'cp_method': cp_utils.analytic_cpmethod()
    }


def rvs(n, x, rust=False, mlcp=True, debug=False):
    """
    Random generation for log-normal with calibrating prior
    
    Parameters
    ----------
    n : int
        Number of samples to generate
    x : array-like
        Training data values
    rust : bool
        Whether to use RUST method (default False)
    mlcp : bool
        Whether to use ML/CP method (default True)
    debug : bool
        Debug flag (default False)
        
    Returns
    -------
    dict
        Dictionary containing random deviates and parameters
    """
    x = cp_utils.to_array(x)
    assert np.all(np.isfinite(x)) and np.all(~np.isnan(x)), "x must be finite and not NaN"
    assert np.all(x >= 0), "x must be non-negative"
    
    ml_params = "mlcp not selected"
    ml_deviates = "mlcp not selected"
    cp_deviates = "mlcp not selected"
    ru_deviates = "rust not selected"
    
    if mlcp:
        q = ppf(x, np.random.uniform(0, 1, n))
        ml_params = q['ml_params']
        ml_deviates = q['ml_quantiles']
        cp_deviates = q['cp_quantiles']
    
    if rust:
        th = tlnorm_cp(n, x)['theta_samples']
        ru_deviates = np.zeros(n)
        for i in range(n):
            ru_deviates[i] = scipy.stats.lognorm.rvs(s=th[i, 1], scale=np.exp(th[i, 0]), size=1)[0]
    
    op = {
        'ml_params': ml_params,
        'ml_deviates': ml_deviates,
        'cp_deviates': cp_deviates,
        'ru_deviates': ru_deviates,
        'cp_method': cp_utils.analytic_cpmethod()
    }
    
    return op


def pdf(x, y=None, rust=False, nrust=1000, debug=False):
    """
    Density function for log-normal with calibrating prior
    
    Parameters
    ----------
    x : array-like
        Training data values
    y : array-like, optional
        Test data values (default: same as x)
    rust : bool
        Whether to use RUST method (default False)
    nrust : int
        Number of RUST samples (default 1000)
    debug : bool
        Debug flag (default False)
        
    Returns
    -------
    dict
        Dictionary containing density values
    """
    x = cp_utils.to_array(x)
    if y is None:
        y = x
    else:
        y = cp_utils.to_array(y)
    
    assert np.all(np.isfinite(x)) and np.all(~np.isnan(x)), "x must be finite and not NaN"
    assert np.all(np.isfinite(y)) and np.all(~np.isnan(y)), "y must be finite and not NaN"
    assert np.all(x >= 0), "x must be non-negative"
    assert np.all(y >= 0), "y must be non-negative"
    
    dd = lnorm_libs.dlnormsub(x=x, y=y)  # dlnormsub expects scalar y
    ru_pdf = "rust not selected"
    
    if rust:
        th = tlnorm_cp(nrust, x)['theta_samples']
        ru_pdf = np.zeros(len(y))
        for ir in range(nrust):
            ru_pdf = ru_pdf + scipy.stats.lognorm.pdf(y, s=th[ir, 1], scale=np.exp(th[ir, 0]))
        ru_pdf = ru_pdf / nrust
    
    op = {
        'ml_params': dd['ml_params'],
        'ml_pdf': dd['ml_pdf'],
        'cp_pdf': dd['rh_pdf'],
        'ru_pdf': ru_pdf,
        'cp_method': cp_utils.analytic_cpmethod()
    }
    
    return op


def cdf(x, y=None, rust=False, nrust=1000, debug=False):
    """
    Cumulative distribution function for log-normal with calibrating prior
    
    Parameters
    ----------
    x : array-like
        Training data values
    y : array-like, optional
        Test data values (default: same as x)
    rust : bool
        Whether to use RUST method (default False)
    nrust : int
        Number of RUST samples (default 1000)
    debug : bool
        Debug flag (default False)
        
    Returns
    -------
    dict
        Dictionary containing CDF values
    """
    x = cp_utils.to_array(x)
    if y is None:
        y = x
    else:
        y = cp_utils.to_array(y)
    
    assert np.all(np.isfinite(x)) and np.all(~np.isnan(x)), "x must be finite and not NaN"
    assert np.all(np.isfinite(y)) and np.all(~np.isnan(y)), "y must be finite and not NaN"
    assert np.all(x >= 0), "x must be non-negative"
    assert np.all(y >= 0), "y must be non-negative"
    
    dd = lnorm_libs.dlnormsub(x=x, y=y[0] if len(y) == 1 else y)  # dlnormsub expects scalar y
    ru_cdf = "rust not selected"
    
    if rust:
        th = tlnorm_cp(nrust, x)['theta_samples']
        ru_cdf = np.zeros(len(y))
        for ir in range(nrust):
            ru_cdf = ru_cdf + scipy.stats.lognorm.cdf(y, s=th[ir, 1], scale=np.exp(th[ir, 0]))
        ru_cdf = ru_cdf / nrust
    
    op = {
        'ml_params': dd['ml_params'],
        'ml_cdf': dd['ml_cdf'],
        'cp_cdf': dd['rh_cdf'],
        'ru_cdf': ru_cdf,
        'cp_method': cp_utils.analytic_cpmethod()
    }
    
    return op


def tlnorm_cp(n, x, debug=False):
    """
    Not yet implemented: Theta sampling for log-normal with calibrating prior
    
    Parameters
    ----------
    n : int
        Number of samples to generate
    x : array-like
        Training data values
    debug : bool
        Debug flag (default False)
        
    Returns
    -------
    dict
        Dictionary containing theta samples
    """
    x = cp_utils.to_array(x)
    assert np.all(np.isfinite(x)) and np.all(~np.isnan(x)), "x must be finite and not NaN"
    assert np.all(x >= 0), "x must be non-negative"
    
    # Initialize with method of moments estimates
    log_x = np.log(x)
    init_meanlog = np.mean(log_x)
    init_sdlog = np.std(log_x, ddof=1)
    
    t = cp_utils.ru(lnorm_libs.lnorm_logf, x=x, n=n, d=2, init=[init_meanlog, init_sdlog])
    
    return {'theta_samples': t['sim_vals']}



import sympy
import autodiff

x, t, v1, v2, v3, v4 = sympy.symbols('x t v1 v2 v3 v4')

f_gev_p1_density = (1/v3)*((1+v4*((x-v1-v2*t)/v3))**(-1/v4-1))*sympy.exp(-((1+v4*(x-v1-v2*t)/v3)**(-1/v4)))
f_gev_p1_cdf = sympy.exp(-((1+v4*(x-v1-v2*t)/v3)**(-1/v4)))
f_gev_p1_log_density = -sympy.log(v3)-(1+1/v4)*sympy.log(1+v4*(x-v1-v2*t)/v3)-(1+v4*(x-v1-v2*t)/v3)**(-1/v4)
                                 
with open('gev_p1_autodiff_output.py', 'w') as f:
    f.write('import numpy as np\n')
    autodiff.diff_to_file(f, f_gev_p1_density, [v1, v2, v3, v4], 'gev_p1_fd', 1, xt=True)
    autodiff.diff_to_file(f, f_gev_p1_density, [v1, v2, v3, v4], 'gev_p1_fdd', 2, xt=True)
    autodiff.diff_to_file(f, f_gev_p1_cdf, [v1, v2, v3, v4], 'gev_p1_pd', 1, xt=True)
    autodiff.diff_to_file(f, f_gev_p1_cdf, [v1, v2, v3, v4], 'gev_p1_pdd', 2, xt=True)
    autodiff.diff_to_file(f, f_gev_p1_log_density, [v1, v2, v3, v4], 'gev_p1_logfdd', 2, xt=True)
    autodiff.diff_to_file(f, f_gev_p1_log_density, [v1, v2, v3, v4], 'gev_p1_logfddd', 3, xt=True)

f.close()


import numpy as np
import scipy.stats as stats
import scipy.optimize as optimize

from . import utils as cp_utils
from . import evaluate_dmgs_equation as cp_dmgs
from . import gamma_libs
from . import gamma_derivs


def ppf(x, p=None, fd1=0.01, fd2=0.01, means=False, waicscores=False, 
              logscores=False, dmgs=True, rust=False, nrust=100000, prior="type 1", 
              debug=False, aderivs=True):
    """
    Gamma Distribution Quantile Predictions Based on a Calibrating Prior
    
    Parameters
    ----------
    x : array-like
        Training data values
    p : array-like, optional
        Probabilities for quantiles (default: seq(0.1,0.9,0.1))
    fd1 : float
        Step size for v1 finite differences (default 0.01)
    fd2 : float  
        Step size for v2 finite differences (default 0.01)
    means : bool
        Whether to calculate means (default False)
    waicscores : bool
        Whether to calculate WAIC scores (default False)
    logscores : bool
        Whether to calculate log scores (default False)
    dmgs : bool
        Whether to use DMGS integration (default True)
    rust : bool
        Whether to use RUST (default False)
    nrust : int
        Number of RUST samples (default 100000)
    prior : str
        Prior type ("type 1" or "type 2", default "type 1")
    debug : bool
        Debug flag (default False)
    aderivs : bool
        Whether to use analytical derivatives (default True)
        
    Returns
    -------
    dict
        Dictionary with ml_params, ml_value, standard_errors, ml_quantiles, 
        cp_quantiles, ru_quantiles, maic, waic1, waic2, ml_oos_logscore,
        cp_oos_logscore, ml_mean, cp_mean, cp_method
    """
    
    # 1 intro
    debug = True
    debug = False
    
    if p is None:
        p = np.arange(0.1, 1.0, 0.1)
    
    x = np.array(x)
    p = np.array(p)
    
    assert np.all(np.isfinite(x)), "x must be finite"
    assert np.all(~np.isnan(x)), "x must not contain NaN"
    assert np.all(np.isfinite(p)), "p must be finite" 
    assert np.all(~np.isnan(p)), "p must not contain NaN"
    assert np.all(p > 0), "p must be > 0"
    assert np.all(p < 1), "p must be < 1"
    assert np.all(x >= 0), "x must be >= 0"
    
    alpha = 1 - p
    nx = len(x)
    nalpha = len(alpha)
    
    # 2 ml param estimate
    opt = optimize.minimize(lambda params: -gamma_libs.gamma_loglik(params, x), 
                          x0=[1, 1], method='BFGS')
    v1hat = opt.x[0]
    v2hat = opt.x[1]
    ml_params = np.array([v1hat, v2hat])
    
    if debug:
        print(f"  v1hat,v2hat={v1hat},{v2hat}")
    
    # 3 aic
    ml_value = -opt.fun
    maic = cp_utils.make_maic(ml_value, nparams=2)
    
    # 4 ml quantiles (vectorized over alpha)
    ml_quantiles = stats.gamma.ppf(1-alpha, a=v1hat, scale=v2hat)
    
    # test of gg code (for future implementation of mpd theory, as a test of the mpd code)
    # gamma_gg(v1hat,fd1,v2hat,fd2)
    
    # 5 dmgs
    standard_errors = "dmgs not selected"
    cp_quantiles = "dmgs not selected"
    ru_quantiles = "dmgs not selected"
    waic1 = "dmgs not selected"
    waic2 = "dmgs not selected"
    ml_oos_logscore = "dmgs not selected"
    cp_oos_logscore = "dmgs not selected"
    cp_oos_logscore = "dmgs not selected"
    ml_mean = "dmgs not selected"
    cp_mean = "dmgs not selected"
    cp_mean = "dmgs not selected"
    cp_method = "dmgs not selected"
    
    if dmgs:
        # 6 lddi
        if aderivs:
            ldd = gamma_derivs.gamma_ldda(x, v1hat, v2hat)
        if not aderivs:
            ldd = gamma_derivs.gamma_ldd(x, v1hat, fd1, v2hat, fd2)
        lddi = np.linalg.inv(ldd)
        standard_errors = cp_utils.make_se(nx, lddi)
        
        # 7 lddd
        if debug:
            print("  calculate lddd")
        if aderivs:
            lddd = gamma_derivs.gamma_lddda(x, v1hat, v2hat)
        if not aderivs:
            lddd = gamma_derivs.gamma_lddd(x, v1hat, fd1, v2hat, fd2)
        
        # 8 mu1
        if debug:
            print("calculate mu1")
        mu1 = gamma_libs.gamma_mu1f(alpha, v1hat, fd1, v2hat, fd2)
        
        # 9 mu2
        if debug:
            print("calculate mu2")
        mu2 = gamma_libs.gamma_mu2f(alpha, v1hat, fd1, v2hat, fd2)
        
        # 10 q cp
        # (I compared these two priors, and the double prior worked better)
        # (so I made the better one type 1)
        # (see the actuary paper)
        if debug:
            print("  cp")
        if prior == "type 1":
            lambdad_cp = np.array([-1/v1hat, -1/v2hat])
        elif prior == "type 2":
            lambdad_cp = np.array([0, -1/v2hat])
        else:
            print("invalid prior choice.")
            raise ValueError("invalid prior choice.")
        
        # lambdad_cp=c(0,-1/v2hat)               #this worked ok
        # lambdad_cp=c(-1/v1hat,-1/v2hat)        #but this worked better
        
        # 11 derive the bayesian dq based on v2hat
        if debug:
            print("  fhat, dq and cp quantiles")
        fhat = stats.gamma.pdf(ml_quantiles, a=v1hat, scale=v2hat)
        dq = cp_dmgs.dmgs(lddi, lddd, mu1, lambdad_cp, mu2, dim=2)
        cp_quantiles = ml_quantiles + dq/(nx*fhat)
        
        # 12 means
        means_result = gamma_libs.gamma_means(means, ml_params, lddi, lddd, lambdad_cp, nx, dim=2)
        ml_mean = means_result['ml_mean']
        cp_mean = means_result['cp_mean']
        
        # 13 waicscores
        waic = gamma_libs.gamma_waic(waicscores, x, v1hat, fd1, v2hat, fd2, lddi, lddd, 
                         lambdad_cp, aderivs)
        waic1 = waic['waic1']
        waic2 = waic['waic2']
        
        # 14 logscores
        logscores_result = gamma_libs.gamma_logscores(logscores, x, fd1, fd2, aderivs)
        ml_oos_logscore = logscores_result['ml_oos_logscore']
        cp_oos_logscore = logscores_result['cp_oos_logscore']
        
        # 15 rust
        ru_quantiles = "rust not selected"
        if rust:
            rustsim = rvs(nrust, x, rust=True, mlcp=False)
            ru_quantiles = cp_utils.makeq(rustsim['ru_deviates'], p)
            
    # end of if dmgs
    
    # return
    return {
        'ml_params': ml_params,
        'ml_value': ml_value,
        # 'ldd': ldd,
        # 'lddi': lddi,
        # 'expinfmat': expinfmat,
        # 'expinfmati': expinfmati,
        'standard_errors': standard_errors,
        'ml_quantiles': ml_quantiles,
        'cp_quantiles': cp_quantiles,
        'ru_quantiles': ru_quantiles,
        'maic': maic,
        'waic1': waic1,
        'waic2': waic2,
        'ml_oos_logscore': ml_oos_logscore,
        'cp_oos_logscore': cp_oos_logscore,
        'ml_mean': ml_mean,
        'cp_mean': cp_mean,
        'cp_method': cp_utils.adhoc_dmgs_cpmethod()
    }


def rvs(n, x, fd1=0.01, fd2=0.01, rust=False, mlcp=True, debug=False, aderivs=True):
    """
    Random number generation for Gamma Distribution with Calibrating Prior
    
    Parameters
    ----------
    n : int
        Number of samples to generate
    x : array-like
        Training data values
    fd1 : float
        Step size for v1 finite differences (default 0.01)
    fd2 : float
        Step size for v2 finite differences (default 0.01)
    rust : bool
        Whether to use RUST sampling (default False)
    mlcp : bool
        Whether to use ML/CP sampling (default True)
    debug : bool
        Debug flag (default False)
    aderivs : bool
        Whether to use analytical derivatives (default True)
        
    Returns
    -------
    dict
        Dictionary with ml_params, ml_deviates, cp_deviates, ru_deviates, cp_method
    """
    
    # stopifnot(is.finite(n),!is.na(n),is.finite(x),!is.na(x),!x<0)
    x = np.array(x)
    assert np.all(np.isfinite(x)), "x must be finite"
    assert np.all(~np.isnan(x)), "x must not contain NaN"
    assert np.all(x >= 0), "x must be >= 0"
    
    ml_params = "mlcp not selected"
    ml_deviates = "mlcp not selected"
    cp_deviates = "mlcp not selected"
    ru_deviates = "rust not selected"
    
    if mlcp:
        q = ppf(x, np.random.uniform(0, 1, n), fd1=fd1, fd2=fd2, aderivs=aderivs)
        ml_params = q['ml_params']
        ml_deviates = q['ml_quantiles']
        cp_deviates = q['cp_quantiles']
    
    if rust:
        th = tgamma_cp(n, x)['theta_samples']
        ru_deviates = np.zeros(n)
        for i in range(n):
            ru_deviates[i] = np.random.gamma(shape=th[i,0], scale=th[i,1])
    
    op = {
        'ml_params': ml_params,
        'ml_deviates': ml_deviates,
        'cp_deviates': cp_deviates,
        'ru_deviates': ru_deviates,
        'cp_method': cp_utils.adhoc_dmgs_cpmethod()
    }
    return op


def pdf(x, y=None, fd1=0.01, fd2=0.01, rust=False, nrust=1000, debug=False, aderivs=True):
    """
    Density function for Gamma Distribution with Calibrating Prior
    
    Parameters
    ----------
    x : array-like
        Training data values
    y : array-like, optional
        Evaluation points (default: same as x)
    fd1 : float
        Step size for v1 finite differences (default 0.01)
    fd2 : float
        Step size for v2 finite differences (default 0.01)
    rust : bool
        Whether to use RUST sampling (default False)
    nrust : int
        Number of RUST samples (default 1000)
    debug : bool
        Debug flag (default False)
    aderivs : bool
        Whether to use analytical derivatives (default True)
        
    Returns
    -------
    dict
        Dictionary with ml_params, ml_pdf, cp_pdf, ru_pdf, cp_method
    """

    x = cp_utils.to_array(x)
    if y is None:
        y = x
    else:
        y = cp_utils.to_array(y)
    
    assert np.all(np.isfinite(x)), "x must be finite"
    assert np.all(~np.isnan(x)), "x must not contain NaN"
    assert np.all(np.isfinite(y)), "y must be finite"
    assert np.all(~np.isnan(y)), "y must not contain NaN"
    assert np.all(x >= 0), "x must be >= 0"
    assert np.all(y >= 0), "y must be >= 0"
    
    dd = gamma_libs.dgammasub(x=x, y=y, fd1=fd1, fd2=fd2, aderivs=aderivs)
    ru_pdf = "rust not selected"
    
    if rust:
        th = tgamma_cp(nrust, x)['theta_samples']
        ru_pdf = np.zeros(len(y))
        for ir in range(nrust):
            ru_pdf = ru_pdf + stats.gamma.pdf(y, a=th[ir,0], scale=th[ir,1])
        ru_pdf = ru_pdf / nrust
    
    op = {
        'ml_params': dd['ml_params'],
        'ml_pdf': dd['ml_pdf'],
        'cp_pdf': dd['cp_pdf'],
        'ru_pdf': ru_pdf,
        'cp_method': cp_utils.adhoc_dmgs_cpmethod()
    }
    return op

def cdf(x, y=None, fd1=0.01, fd2=0.01, rust=False, nrust=1000, debug=False, aderivs=True):
    """
    CDF function for Gamma Distribution with Calibrating Prior
    
    Parameters
    ----------
    x : array-like
        Training data values
    y : array-like, optional
        Evaluation points (default: same as x)
    fd1 : float
        Step size for v1 finite differences (default 0.01)
    fd2 : float
        Step size for v2 finite differences (default 0.01)
    rust : bool
        Whether to use RUST sampling (default False)
    nrust : int
        Number of RUST samples (default 1000)
    debug : bool
        Debug flag (default False)
    aderivs : bool
        Whether to use analytical derivatives (default True)
        
    Returns
    -------
    dict
        Dictionary with ml_params, ml_cdf, cp_cdf, ru_cdf, cp_method
    """
    
    if y is None:
        y = x
    
    x = np.array(x)
    y = np.array(y)
    
    assert np.all(np.isfinite(x)), "x must be finite"
    assert np.all(~np.isnan(x)), "x must not contain NaN"
    assert np.all(np.isfinite(y)), "y must be finite"
    assert np.all(~np.isnan(y)), "y must not contain NaN"
    assert np.all(x >= 0), "x must be >= 0"
    assert np.all(y >= 0), "y must be >= 0"
    
    dd = gamma_libs.dgammasub(x=x, y=y, fd1=fd1, fd2=fd2, aderivs=aderivs)
    ru_cdf = "rust not selected"
    
    if rust:
        th = tgamma_cp(nrust, x)['theta_samples']
        ru_cdf = np.zeros(len(y))
        for ir in range(nrust):
            ru_cdf = ru_cdf + stats.gamma.cdf(y, a=th[ir,0], scale=th[ir,1])
        ru_cdf = ru_cdf / nrust
    
    op = {
        'ml_params': dd['ml_params'],
        'ml_cdf': dd['ml_cdf'],
        'cp_cdf': dd['cp_cdf'],
        'ru_cdf': ru_cdf,
        'cp_method': cp_utils.adhoc_dmgs_cpmethod()
    }
    return op


def tgamma_cp(n, x, fd1=0.01, fd2=0.01, debug=False):
    """
    Not yet implemented: Theta sampling for Gamma Distribution with Calibrating Prior
    
    Parameters
    ----------
    n : int
        Number of theta samples to generate
    x : array-like
        Training data values
    fd1 : float
        Step size for v1 finite differences (default 0.01)
    fd2 : float
        Step size for v2 finite differences (default 0.01)
    debug : bool
        Debug flag (default False)
        
    Returns
    -------
    dict
        Dictionary with theta_samples
    """
    
    x = np.array(x)
    assert np.all(np.isfinite(x)), "x must be finite"
    assert np.all(~np.isnan(x)), "x must not contain NaN"
    assert np.all(x >= 0), "x must be >= 0"
    
    t = cp_utils.ru(gamma_libs.gamma_logf, x=x, n=n, d=2, init=np.array([1, 1]))
    
    return {'theta_samples': t['sim_vals']}


import sympy
import autodiff

x, v1, v2, v3 = sympy.symbols('x v1 v2 v3')

# v1=mu, v2=sigma, v3=xi

f = (1/v2)*(1+v3*((x-v1)/v2))**(-((v3+1)/v3))
p = 1-(1+v3*((x-v1)/v2))**(-1/v3)
logf = -sympy.log(v2)-((1+v3)/v3)*sympy.log(1+v3*(x-v1)/v2)
                                 
with open('output_gpd.py', 'w') as file:
    file.write('import numpy as np\n')
    autodiff.diff_to_file(file, f, [v2, v3], 'gpd_k1_fd', 1)
    autodiff.diff_to_file(file, f, [v2, v3], 'gpd_k1_fdd', 2)
    autodiff.diff_to_file(file, p, [v2, v3], 'gpd_k1_pd', 1)
    autodiff.diff_to_file(file, p, [v2, v3], 'gpd_k1_pdd', 2)
    autodiff.diff_to_file(file, logf, [v2, v3], 'gpd_k1_logfdd', 2)
    autodiff.diff_to_file(file, logf, [v2, v3], 'gpd_k1_logfddd', 3)

file.close()


import numpy as np
import scipy.stats
import scipy.optimize

from . import utils as cp_utils
from . import evaluate_dmgs_equation as cp_dmgs
from . import weibull_libs
from . import weibull_derivs


def ppf(x, p=np.arange(0.1, 1.0, 0.1), means=False, waicscores=False, 
                logscores=False, dmgs=True, rust=False, nrust=100000, debug=False):
    """
    Weibull Distribution Predictions Based on a Calibrating Prior
    
    Parameters
    ----------
    x : array_like
        Data values
    p : array_like, optional
        Probabilities for quantiles (default 0.1 to 0.9 by 0.1)
    means : bool, optional
        Whether to calculate means (default False)
    waicscores : bool, optional
        Whether to calculate WAIC scores (default False)
    logscores : bool, optional
        Whether to calculate log scores (default False)
    dmgs : bool, optional
        Whether to use DMGS integration (default True)
    rust : bool, optional
        Whether to use RUST sampling (default False)
    nrust : int, optional
        Number of RUST samples (default 100000)
    debug : bool, optional
        Whether to print debug messages (default False)
        
    Returns
    -------
    dict
        Dictionary containing results
    """
    # 1 intro
    debug = True
    debug = False
    
    # Input validation
    x = cp_utils.to_array(x)
    p = cp_utils.to_array(p)
    
    assert np.all(np.isfinite(x)) and np.all(~np.isnan(x)), "x must be finite and not NaN"
    assert np.all(np.isfinite(p)) and np.all(~np.isnan(p)), "p must be finite and not NaN"
    assert np.all(p > 0) and np.all(p < 1), "p must be between 0 and 1"
    assert np.all(x >= 0), "x must be non-negative"
    
    alpha = 1 - p
    nx = len(x)
    nalpha = len(alpha)
    
    # 2 ml param estimate
    opt = scipy.optimize.minimize(lambda params: -weibull_libs.weibull_loglik(params, x), 
                                 [1, 1], method='BFGS')
    v1hat, v2hat = opt.x
    ml_params = np.array([v1hat, v2hat])
    
    if debug:
        print(f"  v1hat,v2hat={v1hat},{v2hat}")
    
    # 3 aic
    ml_value = -opt.fun
    maic = cp_utils.make_maic(ml_value, nparams=2)
    
    # 4 ml quantiles (vectorized over alpha)
    ml_quantiles = scipy.stats.weibull_min.ppf(1-alpha, c=v1hat, scale=v2hat)
    
    # 5 dmgs
    standard_errors = "dmgs not selected"
    rh_quantiles = "dmgs not selected"
    ru_quantiles = "dmgs not selected"
    waic1 = "dmgs not selected"
    waic2 = "dmgs not selected"
    ml_oos_logscore = "dmgs not selected"
    rh_oos_logscore = "dmgs not selected"
    cp_oos_logscore = "dmgs not selected"
    ml_mean = "dmgs not selected"
    rh_mean = "dmgs not selected"
    cp_mean = "dmgs not selected"
    cp_method = "dmgs not selected"
    
    if dmgs:
        # 6 lddi
        ldd = weibull_derivs.weibull_ldda(x, v1hat, v2hat)
        lddi = np.linalg.inv(ldd)
        standard_errors = cp_utils.make_se(nx, lddi)
        
        # 7 lddd
        if debug:
            print("  calculate lddd")
        lddd = weibull_derivs.weibull_lddda(x, v1hat, v2hat)
        
        # 7 mu1
        if debug:
            print("calculate mu1")
        mu1 = weibull_derivs.weibull_mu1fa(alpha, v1hat, v2hat)
        
        # 8 mu2
        if debug:
            print("calculate mu2")
        mu2 = weibull_derivs.weibull_mu2fa(alpha, v1hat, v2hat)
        
        # 9 q rhp
        if debug:
            print("  rhp")
        lambdad_rhp = np.array([-1/v1hat, -1/v2hat])
        
        # 10 derive the bayesian dq based on v2hat
        if debug:
            print("  fhat, dq and rhp quantiles")
        fhat = scipy.stats.weibull_min.pdf(ml_quantiles, c=v1hat, scale=v2hat)
        dq = cp_dmgs.dmgs(lddi, lddd, mu1, lambdad_rhp, mu2, dim=2)
        rh_quantiles = ml_quantiles + dq / (nx * fhat)
        
        # 11 means
        means_result = weibull_libs.weibull_means(means, ml_params, lddi, lddd, 
                                                  lambdad_rhp, nx, dim=2)
        ml_mean = means_result['ml_mean']
        rh_mean = means_result['rh_mean']
        
        # 12 waicscores
        waic = weibull_libs.weibull_waic(waicscores, x, v1hat, v2hat, lddi, lddd, 
                                        lambdad_rhp)
        waic1 = waic['waic1']
        waic2 = waic['waic2']
        
        # 13 logscores
        logscores_result = weibull_libs.weibull_logscores(logscores, x)
        ml_oos_logscore = logscores_result['ml_oos_logscore']
        rh_oos_logscore = logscores_result['rh_oos_logscore']
        
        # 14 rust
        ru_quantiles = "rust not selected"
        if rust:
            rustsim = rvs(nrust, x, rust=True, mlcp=False)
            ru_quantiles = cp_utils.makeq(rustsim['ru_deviates'], p)
    
    # return
    return {
        'ml_params': ml_params,
        'ml_value': ml_value,
        'standard_errors': standard_errors,
        'ml_quantiles': ml_quantiles,
        'cp_quantiles': rh_quantiles,
        'ru_quantiles': ru_quantiles,
        'maic': maic,
        'waic1': waic1,
        'waic2': waic2,
        'ml_oos_logscore': ml_oos_logscore,
        'cp_oos_logscore': rh_oos_logscore,
        'ml_mean': ml_mean,
        'cp_mean': rh_mean,
        'cp_method': cp_utils.rhp_dmgs_cpmethod()
    }

def rvs(n, x, rust=False, mlcp=True, debug=False):
    """
    Random generation for Weibull CP
    
    Parameters
    ----------
    n : int
        Number of samples
    x : array_like
        Data for parameter estimation
    rust : bool, optional
        Whether to use RUST sampling (default False)
    mlcp : bool, optional
        Whether to use ML/CP approach (default True)
    debug : bool, optional
        Whether to print debug messages (default False)
        
    Returns
    -------
    dict
        Dictionary containing generated samples
    """
    # Input validation
    x = cp_utils.to_array(x)

    assert np.all(np.isfinite(x)) and np.all(~np.isnan(x)), "x must be finite and not NaN"
    assert np.all(x >= 0), "x must be non-negative"
    
    ml_params = "mlcp not selected"
    ml_deviates = "mlcp not selected"
    cp_deviates = "mlcp not selected"
    ru_deviates = "rust not selected"
    
    if mlcp:
        q = ppf(x, np.random.uniform(0, 1, n))
        ml_params = q['ml_params']
        ml_deviates = q['ml_quantiles']
        cp_deviates = q['cp_quantiles']
    
    if rust:
        th = tweibull_cp(n, x)['theta_samples']
        ru_deviates = np.zeros(n)
        for i in range(n):
            ru_deviates[i] = scipy.stats.weibull_min.rvs(c=th[i, 0], scale=th[i, 1], size=1)
    
    op = {
        'ml_params': ml_params,
        'ml_deviates': ml_deviates,
        'cp_deviates': cp_deviates,
        'ru_deviates': ru_deviates,
        'cp_method': cp_utils.rhp_dmgs_cpmethod()
    }
    return op

def pdf(x, y=None, rust=False, nrust=1000, debug=False):
    """
    Density function for Weibull CP
    
    Parameters
    ----------
    x : array_like
        Data for parameter estimation
    y : array_like, optional
        Points to evaluate density (default same as x)
    rust : bool, optional
        Whether to use RUST sampling (default False)
    nrust : int, optional
        Number of RUST samples (default 1000)
    debug : bool, optional
        Whether to print debug messages (default False)
        
    Returns
    -------
    dict
        Dictionary containing density values
    """

    x = cp_utils.to_array(x)
    if y is None:
        y = x
    else:
        y = cp_utils.to_array(y)
    
    assert np.all(np.isfinite(x)) and np.all(~np.isnan(x)), "x must be finite and not NaN"
    assert np.all(np.isfinite(y)) and np.all(~np.isnan(y)), "y must be finite and not NaN"
    assert np.all(x >= 0), "x must be non-negative"
    assert np.all(y >= 0), "y must be non-negative"
    
    dd = weibull_libs.dweibullsub(x=x, y=y)
    ru_pdf = "rust not selected"
    
    if rust:
        th = tweibull_cp(nrust, x)['theta_samples']
        ru_pdf = np.zeros(len(y))
        for ir in range(nrust):
            ru_pdf += scipy.stats.weibull_min.pdf(y, c=th[ir, 0], scale=th[ir, 1])
        ru_pdf = ru_pdf / nrust
    
    op = {
        'ml_params': dd['ml_params'],
        'ml_pdf': dd['ml_pdf'],
        'cp_pdf': dd['rh_pdf'],
        'ru_pdf': ru_pdf,
        'cp_method': cp_utils.rhp_dmgs_cpmethod()
    }
    return op

def cdf(x, y=None, rust=False, nrust=1000, debug=False):
    """
    Distribution function for Weibull CP
    
    Parameters
    ----------
    x : array_like
        Data for parameter estimation
    y : array_like, optional
        Points to evaluate CDF (default same as x)
    rust : bool, optional
        Whether to use RUST sampling (default False)
    nrust : int, optional
        Number of RUST samples (default 1000)
    debug : bool, optional
        Whether to print debug messages (default False)
        
    Returns
    -------
    dict
        Dictionary containing CDF values
    """
    x = cp_utils.to_array(x)
    if y is None:
        y = x
    else:
        y = cp_utils.to_array(y)

    assert np.all(np.isfinite(x)) and np.all(~np.isnan(x)), "x must be finite and not NaN"
    assert np.all(np.isfinite(y)) and np.all(~np.isnan(y)), "y must be finite and not NaN"
    assert np.all(x >= 0), "x must be non-negative"
    assert np.all(y >= 0), "y must be non-negative"
    
    dd = weibull_libs.dweibullsub(x=x, y=y)
    ru_cdf = "rust not selected"
    
    if rust:
        th = tweibull_cp(nrust, x)['theta_samples']
        ru_cdf = np.zeros(len(y))
        for ir in range(nrust):
            ru_cdf += scipy.stats.weibull_min.cdf(y, c=th[ir, 0], scale=th[ir, 1])
        ru_cdf = ru_cdf / nrust
    
    op = {
        'ml_params': dd['ml_params'],
        'ml_cdf': dd['ml_cdf'],
        'cp_cdf': dd['rh_cdf'],
        'ru_cdf': ru_cdf,
        'cp_method': cp_utils.rhp_dmgs_cpmethod()
    }
    return op

def tweibull_cp(n, x, debug=False):
    """
    Not yet implemented: Theta sampling for Weibull CP
    
    Parameters
    ----------
    n : int
        Number of samples
    x : array_like
        Data values
    debug : bool, optional
        Whether to print debug messages (default False)
        
    Returns
    -------
    dict
        Dictionary containing theta samples
    """
    x = cp_utils.to_array(x)

    assert np.all(np.isfinite(x)) and np.all(~np.isnan(x)), "x must be finite and not NaN"
    assert np.all(x >= 0), "x must be non-negative"
    
    t = cp_utils.ru(weibull_derivs.weibull_logf, x=x, n=n, d=2, init=[1, 1])
    
    return {'theta_samples': t['sim_vals']}




import scipy.stats as sp
import sympy
import autodiff

x, v1, v2 = sympy.symbols('x v1 v2')

f = (1/sympy.sqrt(2*sympy.pi))*(1/v2)*sympy.exp(-(x-v1)**2/(2*v2*v2))
logf = -0.5*sympy.log(2*sympy.pi)-sympy.log(v2)-(x-v1)**2/(2*v2*v2)

                                 
with open('output_normal.py', 'w') as file:
    file.write('import numpy as np\n')
    file.write(autodiff.diff_to_code(f, [v1, v2], 'norm_fd', 1))
    file.write(autodiff.diff_to_code(f, [v1, v2], 'norm_fdd', 2))
    file.write(autodiff.diff_to_code(logf, [v1, v2], 'norm_logfdd', 2))
    file.write(autodiff.diff_to_code(logf, [v1, v2], 'norm_logfddd', 3))

file.close()

# norm_pd and norm_pdd done manually


import numpy as np
from scipy import stats

def lnorm_fd(x, v1, v2): 
	#Automatically generated code.
	dim=2 
	if not (isinstance(x, list) or isinstance(x, np.ndarray)):
		x_all=[x]
	else:
		x_all=x
	result = np.zeros((dim, len(x_all)))
	for i in range(len(x_all)):
		x = x_all[i]
		result[0,i] = -np.sqrt(2)*(2*v1 - 2*np.log(x))*np.exp(-(-v1 + np.log(x))**2/(2*v2**2))/(4*np.sqrt(np.pi)*v2**3*x) 
		result[1,i] = -np.sqrt(2)*np.exp(-(-v1 + np.log(x))**2/(2*v2**2))/(2*np.sqrt(np.pi)*v2**2*x) + np.sqrt(2)*(-v1 + np.log(x))**2*np.exp(-(-v1 + np.log(x))**2/(2*v2**2))/(2*np.sqrt(np.pi)*v2**4*x) 
	if len(x_all) == 1:
		result = result[:,0]
	return result 

def lnorm_fdd(x, v1, v2): 
	#Automatically generated code.
	dim=2 
	if not (isinstance(x, list) or isinstance(x, np.ndarray)):
		x_all=[x]
	else:
		x_all=x
	result = np.zeros((dim,dim, len(x_all)))
	for i in range(len(x_all)):
		x = x_all[i]
		result[0,0,i] = -np.sqrt(2)*(1 - (v1 - np.log(x))**2/v2**2)*np.exp(-(v1 - np.log(x))**2/(2*v2**2))/(2*np.sqrt(np.pi)*v2**3*x)
		result[0,1,i] = np.sqrt(2)*(3 - (v1 - np.log(x))**2/v2**2)*(v1 - np.log(x))*np.exp(-(v1 - np.log(x))**2/(2*v2**2))/(2*np.sqrt(np.pi)*v2**4*x)
		result[1,0,i] = np.sqrt(2)*(3 - (v1 - np.log(x))**2/v2**2)*(v1 - np.log(x))*np.exp(-(v1 - np.log(x))**2/(2*v2**2))/(2*np.sqrt(np.pi)*v2**4*x)
		result[1,1,i] = np.sqrt(2)*(1 - (3 - (v1 - np.log(x))**2/v2**2)*(v1 - np.log(x))**2/(2*v2**2) - (v1 - np.log(x))**2/v2**2)*np.exp(-(v1 - np.log(x))**2/(2*v2**2))/(np.sqrt(np.pi)*v2**3*x)
	if len(x_all) == 1:
		result = result[:,:,0]
	return result



def lnorm_pd(x, v1, v2):    
    e2 = np.log(x) - v1
    e4 = stats.norm.pdf(e2/v2, 0, 1)
    return np.array([-(e4/v2), -(e4 * e2/(v2**2))])


def lnorm_pdd(x, v1, v2):
    e2 = np.log(x) - v1
    e3 = v2**2
    e5 = stats.norm.pdf(e2/v2, 0, 1)
    e7 = e2**2/e3
    e8 = -((e7 - 1) * e5/e3)
    e9 = v2**3
    return np.array([[-(e5 * e2/e9), e8], [e8, -((e7 - 2) * e5 * e2/e9)]])


def lnorm_logfdd(x, v1, v2): 
	#Automatically generated code.
	dim=2 
	if not (isinstance(x, list) or isinstance(x, np.ndarray)):
		x_all=[x]
	else:
		x_all=x
	result = np.zeros((dim,dim, len(x_all)))
	for i in range(len(x_all)):
		x = x_all[i]
		result[0,0,i] = -1/v2**2
		result[0,1,i] = 2*(v1 - np.log(x))/v2**3
		result[1,0,i] = 2*(v1 - np.log(x))/v2**3
		result[1,1,i] = (1 - 3*(v1 - np.log(x))**2/v2**2)/v2**2
	if len(x_all) == 1:
		result = result[:,:,0]
	return result 

def lnorm_logfddd(x, v1, v2): 
	#Automatically generated code.
	dim=2 
	if not (isinstance(x, list) or isinstance(x, np.ndarray)):
		x_all=[x]
	else:
		x_all=x
	result = np.zeros((dim, dim, dim, len(x_all)))
	for i in range(len(x_all)):
		x = x_all[i]
		result[0,0,0,i] = 0 
		result[0,0,1,i] = 2/v2**3 
		result[0,1,0,i] = 2/v2**3 
		result[0,1,1,i] = -6*(v1 - np.log(x))/v2**4 
		result[1,0,0,i] = 2/v2**3 
		result[1,0,1,i] = -6*(v1 - np.log(x))/v2**4 
		result[1,1,0,i] = -6*(v1 - np.log(x))/v2**4 
		result[1,1,1,i] = 2*(-1 + 6*(v1 - np.log(x))**2/v2**2)/v2**3 
	if len(x_all) == 1:
		result = result[:,:,:,0]
	return result 



import numpy as np
import sys
from scipy import stats

from . import utils as cp_utils
from . import norm_derivs


def norm_waic(waicscores, x, v1hat, v2hat):
    """
    Waic
    
    Parameters
    ----------
    waicscores : bool
        Whether to calculate WAIC scores
    x : array_like
        Data values
    v1hat : float
        Estimated parameter 1 (mean)
    v2hat : float
        Estimated parameter 2 (standard deviation)
        
    Returns
    -------
    dict
        Dictionary containing waic1 and waic2 results
    """
    if waicscores:
        f1f = norm_derivs.norm_f1fa(x, v1hat, v2hat)
        f2f = norm_derivs.norm_f2fa(x, v1hat, v2hat)
        ldd = norm_derivs.norm_ldda(x, v1hat, v2hat)
        lddi = np.linalg.inv(ldd)
        lddd = norm_derivs.norm_lddda(x, v1hat, v2hat)
        fhatx = stats.norm.pdf(x, loc=v1hat, scale=v2hat)
        lambdad_rhp = np.array([0, -1/v2hat])
        waic = cp_utils.make_waic(x, fhatx, lddi, lddd, f1f, lambdad_rhp, f2f, dim=2)
        waic1 = waic['waic1']
        waic2 = waic['waic2']
    else:
        waic1 = "waicscores not selected"
        waic2 = "waicscores not selected"
    
    return {'waic1': waic1, 'waic2': waic2}


def norm_logf(params, x):
    """
    Logf for RUST
    
    Parameters
    ----------
    params : array_like
        Parameter vector [mean, std_dev]
    x : array_like
        Data values
        
    Returns
    -------
    float
        Log-likelihood value
    """
    m = params[0]
    s = np.maximum(params[1], sys.float_info.epsilon)
    logf = np.sum(stats.norm.logpdf(x, loc=m, scale=s)) - np.log(s)
    return logf


def norm_ml_params(x):
    """
    Maximum likelihood estimator
    
    Parameters
    ----------
    x : array_like
        Data values
        
    Returns
    -------
    ndarray
        Vector of ML parameter estimates
    """
    mlparams = np.zeros(2)
    nx = len(x)
    mlparams[0] = np.mean(x)
    mlparams[1] = np.sqrt((nx-1)/nx) * np.std(x, ddof=1)
    return mlparams


def norm_unbiasedv_params(x):
    """
    Method of moments estimator
    
    Parameters
    ----------
    x : array_like
        Data values
        
    Returns
    -------
    ndarray
        Vector of parameter estimates
    """
    params = np.zeros(2)
    nx = len(x)
    params[0] = np.mean(x)
    params[1] = np.std(x, ddof=1)
    return params


def norm_logscores(logscores, x):
    """
    Log scores for MLE and RHP predictions calculated using leave-one-out
    
    Parameters
    ----------
    logscores : bool
        Whether to calculate log scores
    x : array_like
        Data values
        
    Returns
    -------
    dict
        Dictionary containing ml_oos_logscore and rh_oos_logscore
    """
    if logscores:
        nx = len(x)
        ml_oos_logscore = 0
        rh_oos_logscore = 0
        
        for i in range(nx):  # 1:nx becomes 0:nx in Python
            x1 = np.concatenate([x[:i], x[i+1:]])  # x[-i] in R becomes this in Python
            
            dd = dnormsub(x1, x[i])
            ml_params = dd['ml_params']
            
            ml_pdf = dd['ml_pdf']
            ml_oos_logscore = ml_oos_logscore + np.log(ml_pdf)
            
            rh_pdf = dd['rh_pdf']
            rh_oos_logscore = rh_oos_logscore + np.log(rh_pdf)
    else:
        ml_oos_logscore = "logscores not selected"
        rh_oos_logscore = "logscores not selected"
    
    return {'ml_oos_logscore': ml_oos_logscore, 'rh_oos_logscore': rh_oos_logscore}


def dnormsub(x, y):
    """
    Densities from MLE and RHP
    
    Parameters
    ----------
    x : array_like
        Training data
    y : float
        Test point
        
    Returns
    -------
    dict
        Dictionary containing ml_params, ml_pdf, rh_pdf, ml_cdf, rh_cdf
    """
    nx = len(x)
    
    # ml
    ml_params = norm_ml_params(x)
    ml_pdf = stats.norm.pdf(y, loc=ml_params[0], scale=ml_params[1])
    ml_cdf = stats.norm.cdf(y, loc=ml_params[0], scale=ml_params[1])
    
    # rhp pdf
    # -first, convert sigma from maxlik to unbiased
    sgu = ml_params[1] * np.sqrt(nx/(nx-1))
    # then, convert sigma to predictive sigma
    sg1 = sgu * np.sqrt((nx+1)/nx)
    
    yd = (y - ml_params[0]) / sg1
    rh_pdf = stats.t.pdf(yd, df=nx-1) / sg1
    rh_pdf = np.maximum(rh_pdf, 0)
    
    # rhp cdf
    rh_cdf = stats.t.cdf(yd, df=nx-1)
    rh_cdf = np.minimum(np.maximum(rh_cdf, 0), 1)
    
    # return
    return {
        'ml_params': ml_params,
        'ml_pdf': ml_pdf,
        'rh_pdf': rh_pdf,
        'ml_cdf': ml_cdf,
        'rh_cdf': rh_cdf
    }



import numpy as np
from scipy import stats, special, optimize
from sklearn.linear_model import LinearRegression
import warnings

from . import genextreme_p12_derivs
from . import utils as cp_utils


def rgev_p12_minmax(nx, mu, sigma, xi, t1, t2, minxi=-0.45, maxxi=0.45, centering=True):
    """
    rgev for gev_p12 but with maxlik xi within bounds
    
    Parameters:
    nx : int
    mu : float
    sigma : float  
    xi : float
    t1 : array-like
    t2 : array-like
    minxi : float, default -0.45
    maxxi : float, default 0.45
    centering : bool, default True
    
    Returns:
    numpy.ndarray : Vector
    """
    xihat = -9999
    
    if centering:
        t1 = t1 - np.mean(t1)
        t2 = t2 - np.mean(t2)
    
    while (xihat < minxi) or (xihat > maxxi):  # 0.46 also works...0.47 doesn't
        # Generate random GEV values using scipy (note: xi -> -xi conversion for scipy)
        xx = stats.genextreme.rvs(c=-xi, loc=mu, scale=sigma, size=nx)
        
        ics = gev_p12_setics(xx, t1, t2, np.array([0, 0, 0, 0, 0]))
        
        # Use scipy.optimize.minimize instead of optim (maximizing = minimizing negative)
        def neg_loglik(params):
            return -gev_p12_loglik(params, xx, t1, t2)
        
        opt1 = optimize.minimize(neg_loglik, ics, method='BFGS')
        xihat = opt1.x[4]
    
    return xx

def gev_p12_waic(waicscores, x, t1, t2, v1hat, v2hat, v3hat, v4hat, v5hat,
                 lddi, lddd, lambdad):
    """
    Waic
    
    Parameters:
    waicscores : bool
    x : array-like
    t1 : array-like
    t2 : array-like
    v1hat : float
    v2hat : float
    v3hat : float
    v4hat : float
    v5hat : float
    lddi : array-like
    lddd : array-like
    lambdad : array-like
    
    Returns:
    dict : Dictionary with waic1 and waic2
    """
    if waicscores:
        f1f = genextreme_p12_derivs.gev_p12_f1fw(x, t1, t2, v1hat, v2hat, v3hat, v4hat, v5hat)
        f2f = genextreme_p12_derivs.gev_p12_f2fw(x, t1, t2, v1hat, v2hat, v3hat, v4hat, v5hat)
        fhatx = dgev_p12(x, t1, t2, ymn=v1hat, slope=v2hat, sigma1=v3hat, sigma2=v4hat, xi=v5hat, log=False)
        waic = cp_utils.make_waic(x, fhatx, lddi, lddd, f1f, lambdad, f2f, dim=5)
        waic1 = waic['waic1']
        waic2 = waic['waic2']
    else:
        waic1 = "waicscores not selected"
        waic2 = "waicscores not selected"
    
    return {'waic1': waic1, 'waic2': waic2}

def gev_p12_predictordata(predictordata, x, t1, t2, t01, t02, params):
    """
    Predicted Parameter and Generalized Residuals
    
    Parameters:
    predictordata : bool
    x : array-like
    t1 : array-like
    t2 : array-like
    t01 : float or array-like
    t02 : float or array-like
    params : array-like
    
    Returns:
    dict : Dictionary with predictedparameter and adjustedx
    """
    if predictordata:
        #
        # calculate the probabilities of the data using the fitted model
        #
        # t is always a matrix
        # t0 is always a vector
        a = params[0]
        b = params[1]
        sc1 = params[2]
        sc2 = params[3]
        sh = params[4]
        
        mu = a + b * t1
        sigma = np.exp(sc1 + sc2 * t2)
        
        # Calculate CDF using scipy genextreme (note: xi -> -xi conversion)
        px = stats.genextreme.cdf(x, c=-sh, loc=mu, scale=sigma)
        
        #
        # calculate the quantiles for those probabilities at t01,t02
        #
        mu0 = a + b * t01
        sigma0 = np.exp(sc1 + sc2 * t02)
        
        # Calculate quantiles using scipy genextreme (note: xi -> -xi conversion)
        qx = stats.genextreme.ppf(px, c=-sh, loc=mu0, scale=sigma0)
        
        predictedparameter = mu
        adjustedx = qx
    else:
        predictedparameter = "predictordata not selected"
        adjustedx = "predictordata not selected"
    
    return {'predictedparameter': predictedparameter, 'adjustedx': adjustedx}

def gev_p12_logf(params, x, t1, t2):
    """
    Logf for RUST
    
    Parameters:
    params : array-like
    x : array-like
    t1 : array-like
    t2 : array-like
    
    Returns:
    float : Log likelihood value
    """
    a = params[0]
    b = params[1]
    sc1 = params[2]
    sc2 = params[3]
    sh = params[4]
    
    # if(is.vector(t)){
    #     mu=a+b*t1
    #     sigma=exp(sc1+sc2*t2)
    # } else {
    #     mu=a+b*t1
    #     sigma=exp(sc1+sc2*t2)
    # }
    
    mu = a + b * t1
    sigma = np.exp(sc1 + sc2 * t2)
    
    # Calculate log PDF using scipy genextreme (note: xi -> -xi conversion)
    logf = np.sum(stats.genextreme.logpdf(x, c=-sh, loc=mu, scale=sigma))
    
    return logf

def gev_p12_setics(x, t1, t2, ics):
    """
    Set initial conditions
    
    Parameters:
    x : array-like
    t1 : array-like
    t2 : array-like
    ics : array-like
    
    Returns:
    numpy.ndarray : Vector
    """
    # t is always a matrix
    nx = len(x)
    
    if (ics[0] == 0) and (ics[1] == 0) and (ics[2] == 0) and (ics[3] == 0) and (ics[4] == 0):
        # Use sklearn LinearRegression instead of R's lm
        reg = LinearRegression()
        t1_reshaped = np.array(t1).reshape(-1, 1)
        reg.fit(t1_reshaped, x)
        
        ics[0] = reg.intercept_
        ics[1] = reg.coef_[0]
        
        xhat = ics[0] + ics[1] * t1
        ics[2] = 0
        ics[3] = 0  # zero because it's inside an exponential
        ics[4] = 0
    
    return ics

def gev_p12_loglik(vv, x, t1, t2):
    """
    observed log-likelihood function
    
    Parameters:
    vv : array-like
    x : array-like
    t1 : array-like
    t2 : array-like
    
    Returns:
    float : Log likelihood value
    """
    # t is always a matrix
    n = len(x)
    
    mu = vv[0] + vv[1] * t1  # so mean is a vector, just like x
    sigma = np.exp(vv[2] + vv[3] * t2)
    
    # Calculate log PDF using scipy genextreme (note: xi -> -xi conversion)
    loglik = np.sum(stats.genextreme.logpdf(x, c=-vv[4], loc=mu, scale=sigma))
    
    return loglik

def gev_p12_checkmle(ml_params, minxi=-1, maxxi=1):
    """
    Check MLE
    
    Parameters:
    ml_params : array-like
    minxi : float, default -1
    maxxi : float, default 1
    
    Returns:
    None : No return value (just a message to the screen).
    """
    # currently not used, because instead I revert2ml
    v1hat = ml_params[0]
    v2hat = ml_params[1]
    v3hat = ml_params[2]
    v4hat = ml_params[3]
    v5hat = ml_params[4]
    
    if np.isnan(v1hat):
        raise RuntimeError()
    if np.isnan(v2hat):
        raise RuntimeError()
    if np.isnan(v3hat):
        raise RuntimeError()
    if np.isnan(v4hat):
        raise RuntimeError()
    if np.isnan(v5hat):
        raise RuntimeError()
        
    if v5hat < minxi:
        warnings.warn(f"\n***v5hat={v5hat}=> execution halted because maxlik shape parameter <{minxi}***")
        raise RuntimeError()
    if v5hat > maxxi:
        warnings.warn(f"\n***v5hat={v5hat}=> execution halted because maxlik shape parameter >{maxxi}***")
        raise RuntimeError()

def qgev_p12(p, t1, t2, ymn, slope, sigma1, sigma2, xi):
    """
    GEVD-with-p1: Quantile function
    
    Parameters:
    p : array-like
    t1 : array-like
    t2 : array-like
    ymn : float
    slope : float
    sigma1 : float
    sigma2 : float
    xi : float
    
    Returns:
    numpy.ndarray : Vector
    """
    # t is sometimes a vector, sometimes a matrix
    
    # if(is.vector(t)){
    #     mu=ymn+slope*t1
    #     sigma=exp(sigma1+sigma2*t2)
    # } else {
    #     mu=ymn+slope*t1
    #     sigma=exp(sigma1+sigma2*t2)
    # }
    
    mu = ymn + slope * t1
    sigma = np.exp(sigma1 + sigma2 * t2)
    
    # Calculate quantiles using scipy genextreme (note: xi -> -xi conversion)
    return stats.genextreme.ppf(p, c=-xi, loc=mu, scale=sigma)

def dgev_p12(x, t1, t2, ymn, slope, sigma1, sigma2, xi, log=False):
    """
    GEVD-with-p1: Density function
    
    Parameters:
    x : array-like
    t1 : array-like
    t2 : array-like
    ymn : float
    slope : float
    sigma1 : float
    sigma2 : float
    xi : float
    log : bool, default False
    
    Returns:
    numpy.ndarray : Vector
    """
    # t is sometimes a vector, sometimes a matrix
    
    # if(is.vector(t)){
    #     mu=ymn+slope*t1
    #     sigma=exp(sigma1+sigma2*t2)
    # } else {
    #     mu=ymn+slope*t1
    #     sigma=exp(sigma1+sigma2*t2)
    # }
    
    mu = ymn + slope * t1
    sigma = np.exp(sigma1 + sigma2 * t2)
    
    # Calculate PDF using scipy genextreme (note: xi -> -xi conversion)
    if log:
        return stats.genextreme.logpdf(x, c=-xi, loc=mu, scale=sigma)
    else:
        return stats.genextreme.pdf(x, c=-xi, loc=mu, scale=sigma)

def pgev_p12(y, t1, t2, ymn, slope, sigma1, sigma2, xi):
    """
    GEVD-with-p1: Distribution function
    
    Parameters:
    y : array-like
    t1 : array-like
    t2 : array-like
    ymn : float
    slope : float
    sigma1 : float
    sigma2 : float
    xi : float
    
    Returns:
    numpy.ndarray : Vector
    """
    # t is sometimes a vector, sometimes a matrix
    
    # if(is.vector(t)){
    #     mu=ymn+slope*t1
    #     sigma=exp(sigma1+sigma2*t2)
    # } else {
    #     mu=ymn+slope*t1
    #     sigma=exp(sigma1+sigma2*t2)
    # }
    
    mu = ymn + slope * t1
    sigma = np.exp(sigma1 + sigma2 * t2)
    
    # Calculate CDF using scipy genextreme (note: xi -> -xi conversion)
    return stats.genextreme.cdf(y, c=-xi, loc=mu, scale=sigma)

def gev_p12_means(means, t01, t02, ml_params, nx):
    """
    Analytical expressions for Predictive Means
    RHP mean based on the expectation of DMGS equation 2.1
    
    Parameters:
    means : bool
    t01 : float or array-like
    t02 : float or array-like
    ml_params : array-like
    nx : int
    
    Returns:
    dict : Dictionary with ml_mean, crhp_mle_mean, and pu_mean
    """
    if means:
        # intro
        eulerconstant = 0.57721566490153286060651209008240243104215933593992
        ymn = ml_params[0]
        slope = ml_params[1]
        sigma1 = ml_params[2]
        sigma2 = ml_params[3]
        sigma = np.exp(sigma1 + sigma2 * t02)
        xi = ml_params[4]
        
        if ml_params[4] == 0:
            # xi=0 case
            ml_mean = ymn + slope * t01 + sigma * eulerconstant
        else:
            # xi!=0 case
            g0 = special.gamma(1 - xi)
            g1 = g0 * special.digamma(1 - xi)
            g2 = (special.polygamma(1, 1 - xi) * g0 * g0 + g1 * g1) / g0
            ml_mean = ymn + slope * t01 + sigma * (g0 - 1) / xi
        
        # return
        crhp_mle_mean = "not yet implemented"
        pu_mean = np.inf
    else:
        pu_mean = "means not selected"
        ml_mean = "means not selected"
        crhp_mle_mean = "means not selected"
    
    return {'ml_mean': ml_mean, 'crhp_mle_mean': crhp_mle_mean, 'pu_mean': pu_mean}

def dgev_p12sub(x, t1, t2, y, t01, t02, ics, minxi, maxxi, debug, extramodels=False):
    """
    Densities for 5 predictions
    
    Parameters:
    x : array-like
    t1 : array-like
    t2 : array-like
    y : array-like
    t01 : float or array-like
    t02 : float or array-like
    ics : array-like
    minxi : float
    maxxi : float
    debug : bool
    extramodels : bool, default False
    
    Returns:
    dict : Dictionary with ml_params, ml_pdf, and ml_cdf
    """
    nx = len(x)
    
    ics = gev_p12_setics(x, t1, t2, ics)
    
    # Use scipy.optimize.minimize instead of optim (maximizing = minimizing negative)
    def neg_loglik(params):
        return -gev_p12_loglik(params, x, t1, t2)
    
    opt = optimize.minimize(neg_loglik, ics, method='BFGS')
    
    v1hat = opt.x[0]
    v2hat = opt.x[1]
    v3hat = opt.x[2]
    v4hat = opt.x[3]
    v5hat = np.minimum(maxxi, np.maximum(minxi, opt.x[4]))
    ml_params = np.array([v1hat, v2hat, v3hat, v4hat, v5hat])
    
    # now that I've dropped dmgs d and p, I don't think I need this any more
    # muhat0=v1hat+v2hat*t01
    # sghat0=exp(v3hat+v4hat*t02)
    # y=fixgevrange(y,muhat0,sghat0,v5hat)
    
    # mle
    ml_pdf = dgev_p12(y, t01, t02, ymn=v1hat, slope=v2hat, sigma1=v3hat, sigma2=v4hat, xi=v5hat)
    ml_cdf = pgev_p12(y, t01, t02, ymn=v1hat, slope=v2hat, sigma1=v3hat, sigma2=v4hat, xi=v5hat)
    
    # return
    return {'ml_params': ml_params, 'ml_pdf': ml_pdf, 'ml_cdf': ml_cdf}




import numpy as np
from scipy.stats import genextreme
from scipy.optimize import minimize

from . import utils as cp_utils
from . import genextreme_p1_libs
from . import genextreme_p12_libs
from . import genextreme_p12_derivs
from . import evaluate_dmgs_equation as cp_dmgs


def ppf(x, t1, t2, t01=np.nan, t02=np.nan, n01=np.nan, n02=np.nan, p=np.arange(0.1, 1.0, 0.1),
                ics=np.array([0, 0, 0, 0, 0]), fdalpha=0.01, minxi=-1, maxxi=1, means=False, waicscores=False,
                extramodels=False, pdf=False, dmgs=True, rust=False, nrust=100000, predictordata=True,
                centering=True, debug=False):
    """
    Generalized Extreme Value Distribution with Two Predictors, Predictions based on a Calibrating Prior
    """
    
    # Input validation
    assert np.all(np.isfinite(x)) and not np.any(np.isnan(x)), "x must be finite and not NA"
    assert np.all(np.isfinite(p)) and not np.any(np.isnan(p)) and np.all(p > 0) and np.all(p < 1), "p must be finite, not NA, and between 0 and 1"
    assert len(t1) == len(x), "t1 must have same length as x"
    assert len(t2) == len(x), "t2 must have same length as x"
    assert len(ics) == 5, "ics must have length 5"
    
    # 1 intro
    alpha = 1 - p
    nx = len(x)
    nalpha = len(alpha)
    t01 = cp_utils.maket0(t01, n01, t1)
    t02 = cp_utils.maket0(t02, n02, t2)
    if debug:
        print(f" t01={t01}")
        print(f" t02={t02}")
    
    if pdf:
        dalpha = np.minimum(fdalpha * alpha, fdalpha * (1 - alpha))
        alpham = alpha - dalpha
        alphap = alpha + dalpha
    
    # 2 centering
    if centering:
        meant1 = np.mean(t1)
        meant2 = np.mean(t2)
        t1 = t1 - meant1
        t2 = t2 - meant2
        t01 = t01 - meant1
        t02 = t02 - meant2
    
    # 3 ml param estimate
    if debug:
        print(" ml param estimate")
    
    # in a small number of cases, direct maxlik fails. I don't really know why, but it gives nonsense
    # so I'm going to try using gev_p1 maxlik first, to determine initial conditions.
    ics = genextreme_p1_libs.gev_p1_setics(x, t1, ics)
    
    def objective1(params):
        return -genextreme_p1_libs.gev_p1_loglik(params, x, t1)
    
    opt1_result = minimize(objective1, ics, method='L-BFGS-B')
    opt1_par = opt1_result.x
    ics[0] = opt1_par[0]
    ics[1] = opt1_par[1]
    ics[2] = np.log(opt1_par[2])
    ics[3] = 0
    ics[4] = 0  # to avoid an initial error occurring sometimes
    
    def objective2(params):
        return -genextreme_p12_libs.gev_p12_loglik(params, x, t1, t2)
    
    opt1_result2 = minimize(objective2, ics, method='L-BFGS-B')  # this one uses the genextreme routine for dgev
    v1hat = opt1_result2.x[0]
    v2hat = opt1_result2.x[1]
    v3hat = opt1_result2.x[2]
    v4hat = opt1_result2.x[3]
    v5hat = opt1_result2.x[4]
    ml_params = np.array([v1hat, v2hat, v3hat, v4hat, v5hat])
    if debug:
        print(f" ml_params={ml_params}")
    
    if abs(v5hat) >= 1:
        revert2ml = True
    else:
        revert2ml = False
    # I'm having some numerical problems with ldd in reliability testing...only in gev_p12...for nx=10 and xi=0.4
    # maybe limiting v5hat to +1 in this way will help
    # for samples of nx=10, and xi=0.4 this will be triggered very frequently, but so be it
    
    # 4 predictordata
    prd = genextreme_p12_libs.gev_p12_predictordata(predictordata, x, t1, t2, t01, t02, ml_params)
    predictedparameter = prd['predictedparameter']
    adjustedx = prd['adjustedx']
    
    # 5 aic
    if debug:
        print(" aic")
    ml_value = -opt1_result2.fun
    maic = cp_utils.make_maic(ml_value, nparams=5)
    
    # 6 calc ml quantiles and density
    if debug:
        print(" ml_quantiles")
    ml_quantiles = genextreme_p12_libs.qgev_p12(1 - alpha, t01, t02, ymn=v1hat, slope=v2hat, sigma1=v3hat, sigma2=v4hat, xi=v5hat)
    if v5hat < 0:
        ml_max = (v1hat + v2hat * t01) - np.exp((v3hat + v4hat * t02)) / v5hat
    else:
        ml_max = np.inf
    fhat = genextreme_p12_libs.dgev_p12(ml_quantiles, t01, t02, ymn=v1hat, slope=v2hat, sigma1=v3hat, sigma2=v4hat, xi=v5hat, log=False)
    if debug:
        print(f" 1: ml_quantiles={ml_quantiles}")
        print(f" 1: fhat={fhat}")
    
    # dmgs
    standard_errors = "dmgs not selected"
    cp_quantiles = "dmgs not selected"
    ru_quantiles = "dmgs not selected"
    ml_pdf = "dmgs not selected"
    cp_pdf = "dmgs not selected"
    rh_flat_pdf = "dmgs not selected"
    waic1 = "dmgs not selected"
    waic2 = "dmgs not selected"
    ml_mean = "dmgs not selected"
    cp_mean = "dmgs not selected"
    rh_flat_mean = "dmgs not selected"
    
    if dmgs and not revert2ml:
        # 7 alpha pdf stuff
        if pdf:
            ml_quantilesm = genextreme_p12_libs.qgev_p12(1 - alpham, t01, t02, ymn=v1hat, slope=v2hat, sigma1=v3hat, sigma2=v4hat, xi=v5hat)
            ml_quantilesp = genextreme_p12_libs.qgev_p12(1 - alphap, t01, t02, ymn=v1hat, slope=v2hat, sigma1=v3hat, sigma2=v4hat, xi=v5hat)
            fhatm = genextreme_p12_libs.dgev_p12(ml_quantilesm, t01, t02, ymn=v1hat, slope=v2hat, sigma1=v3hat, sigma2=v4hat, xi=v5hat, log=False)
            fhatp = genextreme_p12_libs.dgev_p12(ml_quantilesp, t01, t02, ymn=v1hat, slope=v2hat, sigma1=v3hat, sigma2=v4hat, xi=v5hat, log=False)
        
        # 8 ldd
        if debug:
            print("calc ldd")
        ldd = genextreme_p12_derivs.gev_p12_ldda(x, t1, t2, v1hat, v2hat, v3hat, v4hat, v5hat)
        lddi = np.linalg.solve(ldd, np.eye(5))
        standard_errors = cp_utils.make_se(nx, lddi)
        
        # 10 calculate lddd
        if debug:
            print(" calc lddd")
        lddd = genextreme_p12_derivs.gev_p12_lddda(x, t1, t2, v1hat, v2hat, v3hat, v4hat, v5hat)
        
        # 11 mu1
        if debug:
            print(" calculate mu1")
        mu1 = genextreme_p12_derivs.gev_p12_mu1fa(alpha, t01, t02, v1hat, v2hat, v3hat, v4hat, v5hat)
        if pdf:
            mu1m = genextreme_p12_derivs.gev_p12_mu1fa(alpham, t01, t02, v1hat, v2hat, v3hat, v4hat, v5hat)
            mu1p = genextreme_p12_derivs.gev_p12_mu1fa(alphap, t01, t02, v1hat, v2hat, v3hat, v4hat, v5hat)
        
        # 12 mu2
        if debug:
            print(" calculate mu2")
        mu2 = genextreme_p12_derivs.gev_p12_mu2fa(alpha, t01, t02, v1hat, v2hat, v3hat, v4hat, v5hat)
        if pdf:
            if debug:
                print(" alpha pdf option")
            mu2m = genextreme_p12_derivs.gev_p12_mu2fa(alpham, t01, t02, v1hat, v2hat, v3hat, v4hat, v5hat)
            mu2p = genextreme_p12_derivs.gev_p12_mu2fa(alphap, t01, t02, v1hat, v2hat, v3hat, v4hat, v5hat)
        
        # 13 rh_flat model
        if debug:
            print("call dmgs")
        lambdad_cp = np.zeros(5)
        dq = cp_dmgs.dmgs(lddi, lddd, mu1, lambdad_cp, mu2, dim=5)
        if debug:
            print("make cp quantiles")
        rh_flat_quantiles = ml_quantiles + dq / (nx * fhat)
        
        # 15 alpha pdf
        if debug:
            print("step 15")
        if pdf:
            lambdad_crhp_mle = np.zeros(4)
            dqm = cp_dmgs.dmgs(lddi, lddd, mu1m, lambdad_cp, mu2m, dim=5)
            dqp = cp_dmgs.dmgs(lddi, lddd, mu1p, lambdad_cp, mu2p, dim=5)
            quantilesm = ml_quantilesm + dqm / (nx * fhatm)
            quantilesp = ml_quantilesp + dqp / (nx * fhatp)
            ml_pdf = fhat
            rh_flat_pdf = -(alphap - alpham) / (quantilesp - quantilesm)
        else:
            ml_pdf = fhat
            rh_flat_pdf = "pdf not selected"
        
        # 16 means
        if debug:
            print("step 16")
        means_result = genextreme_p12_libs.gev_p12_means(means, t01, t02, ml_params, nx)
        ml_mean = means_result['ml_mean']
        rh_flat_mean = means_result['crhp_mle_mean']
        
        # 17 waicscores
        if debug:
            print("step 17")
        waic = genextreme_p12_libs.gev_p12_waic(waicscores, x, t1, t2, v1hat, v2hat, v3hat, v4hat, v5hat,
                           lddi, lddd, lambdad_cp)
        waic1 = waic['waic1']
        waic2 = waic['waic2']
        
        # 19 rust
        if debug:
            print("step 19")
        ru_quantiles = "rust not selected"
        if rust:
            rustsim = rvs(nrust, x, t1=t1, t2=t2, t01=t01, t02=t02, rust=True, mlcp=False, debug=debug)
            ru_quantiles = cp_utils.makeq(rustsim['ru_deviates'], p)
    else:
        rh_flat_quantiles = ml_quantiles
        ru_quantiles = ml_quantiles
        rh_flat_pdf = ml_pdf
        rh_flat_mean = ml_mean
    # end of if(dmgs)
    
    # 20 decentering
    if debug:
        print("step 20")
    if centering:
        if debug:
            print(f" qgev:ml_params,meant1={ml_params},{meant1}")
        ml_params[0] = ml_params[0] - ml_params[1] * meant1
        ml_params[2] = ml_params[2] - ml_params[3] * meant2
        if predictordata:
            predictedparameter = predictedparameter - ml_params[1] * meant1
    
    if debug:
        print("step 21")
    
    return {
        'ml_params': ml_params,
        'ml_value': ml_value,
        'predictedparameter': predictedparameter,
        'adjustedx': adjustedx,
        'standard_errors': standard_errors,
        'revert2ml': revert2ml,
        'ml_quantiles': ml_quantiles,
        'ml_max': ml_max,
        'cp_quantiles': rh_flat_quantiles,
        'ru_quantiles': ru_quantiles,
        'ml_pdf': ml_pdf,
        'cp_pdf': rh_flat_pdf,
        'maic': maic,
        'waic1': waic1,
        'waic2': waic2,
        'ml_mean': ml_mean,
        'cp_mean': rh_flat_mean,
        'cp_method': cp_utils.crhpflat_dmgs_cpmethod()
    }

def rvs(n, x, t1, t2, t01=np.nan, t02=np.nan, n01=np.nan, n02=np.nan, ics=np.array([0, 0, 0, 0, 0]),
                minxi=-1, maxxi=1, extramodels=False, rust=False, mlcp=True, centering=True, debug=False):
    """Random generation for GEV with two predictors and calibrating prior"""
    
    # Input validation
    assert np.all(np.isfinite(x)) and not np.any(np.isnan(x)), "x must be finite and not NA"
    assert len(t1) == len(x), "t1 must have same length as x"
    assert len(t2) == len(x), "t2 must have same length as x"
    assert np.all(np.isfinite(t1)) and np.all(np.isfinite(t2)) and not np.any(np.isnan(t1)) and not np.any(np.isnan(t2)), "t1 and t2 must be finite and not NA"
    assert len(ics) == 5, "ics must have length 5"
    
    t01 = cp_utils.maket0(t01, n01, t1)
    t02 = cp_utils.maket0(t02, n02, t2)
    
    # centering
    if centering:
        meant1 = np.mean(t1)
        meant2 = np.mean(t2)
        t1 = t1 - meant1
        t2 = t2 - meant2
        t01 = t01 - meant1
        t02 = t02 - meant2
    
    ml_params = "mlcp not selected"
    ml_deviates = "mlcp not selected"
    cp_deviates = "mlcp not selected"
    ru_deviates = "rust not selected"
    
    if mlcp:
        q = ppf(x, t1=t1, t2=t2, t01=t01, t02=t02, n01=np.nan, n02=np.nan, p=np.random.uniform(size=n),
                       ics=ics, extramodels=extramodels, centering=centering)
        ml_params = q['ml_params']
        if debug:
            print(f" inside rgev_p12_cp: ml_params={ml_params}")
        ml_deviates = q['ml_quantiles']
        ru_deviates = q['ru_quantiles']
        cp_deviates = q['cp_quantiles']
    
    if rust:
        th = tgev_p12_cp(n, x, t1, t2)['theta_samples']
        ru_deviates = np.zeros(n)
        for i in range(n):
            mu = th[i, 0] + t01 * th[i, 1]
            sigma = np.exp(th[i, 2] + t02 * th[i, 3])
            xi = th[i, 4]
            ru_deviates[i] = genextreme.rvs(c=-xi, loc=mu, scale=sigma, size=1)
    
    # decentering
    if debug:
        print(f" rgev:ml_params,meant1={ml_params},{meant1}")
    if mlcp and centering:
        ml_params[0] = ml_params[0] - ml_params[1] * meant1
        ml_params[2] = ml_params[2] - ml_params[3] * meant2
    
    op = {
        'ml_params': ml_params,
        'ml_deviates': ml_deviates,
        'cp_deviates': cp_deviates,
        'ru_deviates': ru_deviates,
        'cp_method': cp_utils.crhpflat_dmgs_cpmethod()
    }
    
    return op

def pdf(x, t1, t2, t01=np.nan, t02=np.nan, n01=np.nan, n02=np.nan, y=None, ics=np.array([0, 0, 0, 0, 0]),
                minxi=-1, maxxi=1, extramodels=False, rust=False, nrust=10, centering=True, debug=False):
    """Density function for GEV with two predictors and calibrating prior"""
    
    if y is None:
        y = x
    
    # Input validation
    assert np.all(np.isfinite(x)) and not np.any(np.isnan(x)), "x must be finite and not NA"
    assert np.all(np.isfinite(y)) and not np.any(np.isnan(y)), "y must be finite and not NA"
    assert len(t1) == len(x), "t1 must have same length as x"
    assert len(t2) == len(x), "t2 must have same length as x"
    assert np.all(np.isfinite(t1)) and np.all(np.isfinite(t2)) and not np.any(np.isnan(t1)) and not np.any(np.isnan(t2)), "t1 and t2 must be finite and not NA"
    assert len(ics) == 5, "ics must have length 5"
    
    if debug:
        print(" maket0")
    t01 = cp_utils.maket0(t01, n01, t1)
    t02 = cp_utils.maket0(t02, n02, t2)
    
    # centering
    if centering:
        if debug:
            print(" centering")
        meant1 = np.mean(t1)
        meant2 = np.mean(t2)
        t1 = t1 - meant1
        t2 = t2 - meant2
        t01 = t01 - meant1
        t02 = t02 - meant2
    
    if debug:
        print(" ics and optim")
    ics = genextreme_p12_libs.gev_p12_setics(x, t1, t2, ics)
    
    def objective(params):
        return -genextreme_p12_libs.gev_p12_loglik(params, x, t1, t2)
    
    opt1_result = minimize(objective, ics, method='L-BFGS-B')
    v1hat = opt1_result.x[0]
    v2hat = opt1_result.x[1]
    v3hat = opt1_result.x[2]
    v4hat = opt1_result.x[3]
    v5hat = opt1_result.x[4]
    
    if v5hat <= -1:
        revert2ml = True
    else:
        revert2ml = False
    
    ml_params = np.array([v1hat, v2hat, v3hat, v4hat, v5hat])
    
    if debug:
        print(" call sub")
    dd = genextreme_p12_libs.dgev_p12sub(x=x, t1=t1, t2=t2, y=y, t01=t01, t02=t02, ics=ics,
                    minxi=minxi, maxxi=maxxi, extramodels=extramodels, debug=debug)
    ru_pdf = "rust not selected"
    ml_params = dd['ml_params']
    
    if rust and not revert2ml:
        if debug:
            print(" rust")
        th = tgev_p12_cp(nrust, x=x, t1=t1, t2=t2, debug=debug)['theta_samples']
        if debug:
            print(" tgev call done")
        ru_pdf = np.zeros(len(y))
        for ir in range(nrust):
            mu = th[ir, 0] + t01 * th[ir, 1]
            sigma = np.exp(th[ir, 2] + t02 * th[ir, 3])
            ru_pdf = ru_pdf + genextreme.pdf(y, c=-th[ir, 4], loc=mu, scale=sigma)

        ru_pdf = ru_pdf / nrust
    else:
        ru_pdf = dd['ml_pdf']
    
    # decentering
    if centering:
        if debug:
            print(" decentering")
            print(f" dgev:ml_params,meant1={ml_params},{meant1}")
        ml_params[0] = ml_params[0] - ml_params[1] * meant1
        ml_params[2] = ml_params[2] - ml_params[3] * meant2
    
    op = {
        'ml_params': ml_params,
        'ml_pdf': dd['ml_pdf'],
        'revert2ml': revert2ml,
        'ru_pdf': ru_pdf,
        'cp_method': cp_utils.nopdfcdfmsg()
    }
    
    return op

def cdf(x, t1, t2, t01=np.nan, t02=np.nan, n01=np.nan, n02=np.nan, y=None, ics=np.array([0, 0, 0, 0, 0]),
                minxi=-1, maxxi=1, extramodels=False, rust=False, nrust=1000, centering=True, debug=False):
    """CDF function for GEV with two predictors and calibrating prior"""
    
    if y is None:
        y = x
    
    # Input validation
    assert np.all(np.isfinite(x)) and not np.any(np.isnan(x)), "x must be finite and not NA"
    assert np.all(np.isfinite(y)) and not np.any(np.isnan(y)), "y must be finite and not NA"
    assert len(t1) == len(x), "t1 must have same length as x"
    assert len(t2) == len(x), "t2 must have same length as x"
    assert np.all(np.isfinite(t1)) and np.all(np.isfinite(t2)) and not np.any(np.isnan(t1)) and not np.any(np.isnan(t2)), "t1 and t2 must be finite and not NA"
    assert len(ics) == 5, "ics must have length 5"
    
    t01 = cp_utils.maket0(t01, n01, t1)
    t02 = cp_utils.maket0(t02, n02, t2)
    
    # centering
    if centering:
        meant1 = np.mean(t1)
        meant2 = np.mean(t2)
        t1 = t1 - meant1
        t2 = t2 - meant2
        t01 = t01 - meant1
        t02 = t02 - meant2
    
    ics = genextreme_p12_libs.gev_p12_setics(x, t1, t2, ics)
    
    def objective(params):
        return -genextreme_p12_libs.gev_p12_loglik(params, x, t1, t2)
    
    opt1_result = minimize(objective, ics, method='L-BFGS-B')  # this one uses the genextreme routine for dgev
    v1hat = opt1_result.x[0]
    v2hat = opt1_result.x[1]
    v3hat = opt1_result.x[2]
    v4hat = opt1_result.x[3]
    v5hat = opt1_result.x[4]
    
    if v5hat <= -1:
        revert2ml = True
    else:
        revert2ml = False
    
    ml_params = np.array([v1hat, v2hat, v3hat, v4hat, v5hat])
    
    dd = genextreme_p12_libs.dgev_p12sub(x=x, t1=t1, t2=t2, y=y, t01=t01, t02=t02, ics=ics,
                    minxi=minxi, maxxi=maxxi, extramodels=extramodels, debug=debug)
    ru_cdf = "rust not selected"
    ml_params = dd['ml_params']
    
    if rust and not revert2ml:
        th = tgev_p12_cp(nrust, x, t1, t2)['theta_samples']
        ru_cdf = np.zeros(len(y))
        for ir in range(nrust):
            mu = th[ir, 0] + t01 * th[ir, 1]
            sigma = np.exp(th[ir, 2] + t02 * th[ir, 3])
            xi = th[ir, 4]
            ru_cdf = ru_cdf + genextreme.cdf(y, c=-xi, loc=mu, scale=np.exp(sigma))
            raise Warning(" why is it exp(sigma) here? does it work?")
        ru_cdf = ru_cdf / nrust
    else:
        ru_pdf = dd['ml_pdf']
    
    # decentering
    if centering:
        if debug:
            print(f" pgev:ml_params,meant1={ml_params},{meant1}")
        ml_params[0] = ml_params[0] - ml_params[1] * meant1
        ml_params[2] = ml_params[2] - ml_params[3] * meant2
    
    op = {
        'ml_params': ml_params,
        'ml_cdf': dd['ml_cdf'],
        'revert2ml': revert2ml,
        'ru_cdf': ru_cdf,
        'cp_method': cp_utils.nopdfcdfmsg()
    }
    
    return op

def tgev_p12_cp(n, x, t1, t2, ics=np.array([0, 0, 0, 0, 0]), extramodels=False, debug=False):
    """not implemented: Parameter sampling for GEV with two predictors and calibrating prior"""

    raise Exception('t sampling is not yet implemented.')
    
    # Input validation
    assert np.all(np.isfinite(x)) and not np.any(np.isnan(x)), "x must be finite and not NA"
    assert len(t1) == len(x), "t1 must have same length as x"
    assert len(t2) == len(x), "t2 must have same length as x"
    assert np.all(np.isfinite(t1)) and np.all(np.isfinite(t2)) and not np.any(np.isnan(t1)) and not np.any(np.isnan(t2)), "t1 and t2 must be finite and not NA"
    assert len(ics) == 5, "ics must have length 5"
    
    # centering
    meant1 = np.mean(t1)
    meant2 = np.mean(t2)
    t1 = t1 - meant1
    t2 = t2 - meant2
    
    if debug:
        print(f"sums={np.sum(x)},{np.sum(t1)},{np.sum(t2)},{n}")
    
    th = ru(genextreme_p12_derivs.gev_p12_logf, x=x, t1=t1, t2=t2, n=n, d=5, init=np.array([0, 0, 0, 0, 0]))
    if debug:
        print(" back from rust")
    
    theta_samples = th['sim_vals']
    
    # decentering
    theta_samples[:, 0] = theta_samples[:, 0] - theta_samples[:, 1] * meant1
    theta_samples[:, 2] = theta_samples[:, 2] - theta_samples[:, 3] * meant2
    
    return {'theta_samples': theta_samples}




import numpy as np

def weibull_fd(x, v1, v2): 
	#Automatically generated code.
	dim=2 
	if not (isinstance(x, list) or isinstance(x, np.ndarray)):
		x_all=[x]
	else:
		x_all=x
	result = np.zeros((dim, len(x_all)))
	for i in range(len(x_all)):
		x = x_all[i]
		result[0,i] = -v1*(x/v2)**v1*(x/v2)**(v1 - 1)*np.exp(-(x/v2)**v1)*np.log(x/v2)/v2 + v1*(x/v2)**(v1 - 1)*np.exp(-(x/v2)**v1)*np.log(x/v2)/v2 + (x/v2)**(v1 - 1)*np.exp(-(x/v2)**v1)/v2 
		result[1,i] = v1**2*(x/v2)**v1*(x/v2)**(v1 - 1)*np.exp(-(x/v2)**v1)/v2**2 - v1*(x/v2)**(v1 - 1)*(v1 - 1)*np.exp(-(x/v2)**v1)/v2**2 - v1*(x/v2)**(v1 - 1)*np.exp(-(x/v2)**v1)/v2**2 
	if len(x_all) == 1:
		result = result[:,0]
	return result 

def weibull_fdd(x, v1, v2): 
	#Automatically generated code.
	dim=2 
	if not (isinstance(x, list) or isinstance(x, np.ndarray)):
		x_all=[x]
	else:
		x_all=x
	result = np.zeros((dim,dim, len(x_all)))
	for i in range(len(x_all)):
		x = x_all[i]
		result[0,0,i] = (x/v2)**(v1 - 1)*(v1*(x/v2)**v1*((x/v2)**v1 - 1)*np.log(x/v2) - 2*v1*(x/v2)**v1*np.log(x/v2) + v1*np.log(x/v2) - 2*(x/v2)**v1 + 2)*np.exp(-(x/v2)**v1)*np.log(x/v2)/v2
		result[0,1,i] = (x/v2)**(v1 - 1)*(-v1**2*(x/v2)**(2*v1)*np.log(x/v2) + 2*v1**2*(x/v2)**v1*np.log(x/v2) + v1*(x/v2)**v1*(v1 - 1)*np.log(x/v2) + v1*(x/v2)**v1*np.log(x/v2) + 2*v1*(x/v2)**v1 - v1*(v1 - 1)*np.log(x/v2) - v1*np.log(x/v2) - 2*v1)*np.exp(-(x/v2)**v1)/v2**2
		result[1,0,i] = (x/v2)**(v1 - 1)*(-v1**2*(x/v2)**(2*v1)*np.log(x/v2) + 2*v1**2*(x/v2)**v1*np.log(x/v2) + v1*(x/v2)**v1*(v1 - 1)*np.log(x/v2) + v1*(x/v2)**v1*np.log(x/v2) + 2*v1*(x/v2)**v1 - v1*(v1 - 1)*np.log(x/v2) - v1*np.log(x/v2) - 2*v1)*np.exp(-(x/v2)**v1)/v2**2
		result[1,1,i] = v1*(x/v2)**(v1 - 1)*(-2*v1*(x/v2)**v1*(v1 - 1) - v1*(x/v2)**v1*(-v1*(x/v2)**v1 + v1 + 1) - 2*v1*(x/v2)**v1 + v1*(v1 - 1) + 2*v1)*np.exp(-(x/v2)**v1)/v2**3
	if len(x_all) == 1:
		result = result[:,:,0]
	return result 

def weibull_pd(x, v1, v2): 
	#Automatically generated code.
	dim=2 
	if not (isinstance(x, list) or isinstance(x, np.ndarray)):
		x_all=[x]
	else:
		x_all=x
	result = np.zeros((dim, len(x_all)))
	for i in range(len(x_all)):
		x = x_all[i]
		result[0,i] = (x/v2)**v1*np.exp(-(x/v2)**v1)*np.log(x/v2) 
		result[1,i] = -v1*(x/v2)**v1*np.exp(-(x/v2)**v1)/v2 
	if len(x_all) == 1:
		result = result[:,0]
	return result 

def weibull_pdd(x, v1, v2): 
	#Automatically generated code.
	dim=2 
	if not (isinstance(x, list) or isinstance(x, np.ndarray)):
		x_all=[x]
	else:
		x_all=x
	result = np.zeros((dim,dim, len(x_all)))
	for i in range(len(x_all)):
		x = x_all[i]
		result[0,0,i] = (x/v2)**v1*(1 - (x/v2)**v1)*np.exp(-(x/v2)**v1)*np.log(x/v2)**2
		result[0,1,i] = (x/v2)**v1*(v1*(x/v2)**v1*np.log(x/v2) - v1*np.log(x/v2) - 1)*np.exp(-(x/v2)**v1)/v2
		result[1,0,i] = (x/v2)**v1*(v1*(x/v2)**v1*np.log(x/v2) - v1*np.log(x/v2) - 1)*np.exp(-(x/v2)**v1)/v2
		result[1,1,i] = v1*(x/v2)**v1*(-v1*(x/v2)**v1 + v1 + 1)*np.exp(-(x/v2)**v1)/v2**2
	if len(x_all) == 1:
		result = result[:,:,0]
	return result 

def weibull_logfdd(x, v1, v2): 
	#Automatically generated code.
	dim=2 
	if not (isinstance(x, list) or isinstance(x, np.ndarray)):
		x_all=[x]
	else:
		x_all=x
	result = np.zeros((dim,dim, len(x_all)))
	for i in range(len(x_all)):
		x = x_all[i]
		result[0,0,i] = -((x/v2)**v1*np.log(x/v2)**2 + v1**(-2))
		result[0,1,i] = (v1*(x/v2)**v1*np.log(x/v2) + (x/v2)**v1 - 1)/v2
		result[1,0,i] = (v1*(x/v2)**v1*np.log(x/v2) + (x/v2)**v1 - 1)/v2
		result[1,1,i] = (-v1**2*(x/v2)**v1 - v1*(x/v2)**v1 + v1)/v2**2
	if len(x_all) == 1:
		result = result[:,:,0]
	return result 

def weibull_logfddd(x, v1, v2): 
	#Automatically generated code.
	dim=2 
	if not (isinstance(x, list) or isinstance(x, np.ndarray)):
		x_all=[x]
	else:
		x_all=x
	result = np.zeros((dim, dim, dim, len(x_all)))
	for i in range(len(x_all)):
		x = x_all[i]
		result[0,0,0,i] = -(x/v2)**v1*np.log(x/v2)**3 + 2/v1**3 
		result[0,0,1,i] = (x/v2)**v1*(v1*np.log(x/v2) + 2)*np.log(x/v2)/v2 
		result[0,1,0,i] = (x/v2)**v1*(v1*np.log(x/v2) + 2)*np.log(x/v2)/v2 
		result[0,1,1,i] = (-v1**2*(x/v2)**v1*np.log(x/v2) - v1*(x/v2)**v1*np.log(x/v2) - 2*v1*(x/v2)**v1 - (x/v2)**v1 + 1)/v2**2 
		result[1,0,0,i] = (x/v2)**v1*(v1*np.log(x/v2) + 2)*np.log(x/v2)/v2 
		result[1,0,1,i] = (-v1**2*(x/v2)**v1*np.log(x/v2) - v1*(x/v2)**v1*np.log(x/v2) - 2*v1*(x/v2)**v1 - (x/v2)**v1 + 1)/v2**2 
		result[1,1,0,i] = (-v1**2*(x/v2)**v1*np.log(x/v2) - v1*(x/v2)**v1*np.log(x/v2) - 2*v1*(x/v2)**v1 - (x/v2)**v1 + 1)/v2**2 
		result[1,1,1,i] = (v1**3*(x/v2)**v1 + 3*v1**2*(x/v2)**v1 + 2*v1*(x/v2)**v1 - 2*v1)/v2**3 
	if len(x_all) == 1:
		result = result[:,:,:,0]
	return result 



import numpy as np


def deriv_1(x, t, v1, v2, v3): 
	#Automatically generated code.
	dim=3 
	t_is_vector = isinstance(t, list) or isinstance(t, np.ndarray)
	if t_is_vector:
		t_all = t
	if not (isinstance(x, list) or isinstance(x, np.ndarray)):
		x_all=[x]
	else:
		x_all=x
	result = np.zeros((dim, len(x_all)))
	for i in range(len(x_all)):
		x = x_all[i]
		if t_is_vector:
			t = t_all[i]
		result[0,i] = 3*v1**2*x + 3*v2**2*v3*x 
		result[1,i] = t + 6*v1*v2*v3*x 
		result[2,i] = 3*v1*v2**2*x 
	if len(x_all) == 1:
		result = result[:,0]
	return result 

def deriv_2(x, t, v1, v2, v3): 
	#Automatically generated code.
	dim=3 
	t_is_vector = isinstance(t, list) or isinstance(t, np.ndarray)
	if t_is_vector:
		t_all = t
	if not (isinstance(x, list) or isinstance(x, np.ndarray)):
		x_all=[x]
	else:
		x_all=x
	result = np.zeros((dim,dim, len(x_all)))
	for i in range(len(x_all)):
		x = x_all[i]
		if t_is_vector:
			t = t_all[i]
		result[0,0,i] = 6*v1*x
		result[0,1,i] = 6*v2*v3*x
		result[0,2,i] = 3*v2**2*x
		result[1,0,i] = 6*v2*v3*x
		result[1,1,i] = 6*v1*v3*x
		result[1,2,i] = 6*v1*v2*x
		result[2,0,i] = 3*v2**2*x
		result[2,1,i] = 6*v1*v2*x
		result[2,2,i] = 0
	if len(x_all) == 1:
		result = result[:,:,0]
	return result 

def deriv_3(x, t, v1, v2, v3): 
	#Automatically generated code.
	dim=3 
	t_is_vector = isinstance(t, list) or isinstance(t, np.ndarray)
	if t_is_vector:
		t_all = t
	if not (isinstance(x, list) or isinstance(x, np.ndarray)):
		x_all=[x]
	else:
		x_all=x
	result = np.zeros((dim, dim, dim, len(x_all)))
	for i in range(len(x_all)):
		x = x_all[i]
		if t_is_vector:
			t = t_all[i]
		result[0,0,0,i] = 6*x 
		result[0,0,1,i] = 0 
		result[0,0,2,i] = 0 
		result[0,1,0,i] = 0 
		result[0,1,1,i] = 6*v3*x 
		result[0,1,2,i] = 6*v2*x 
		result[0,2,0,i] = 0 
		result[0,2,1,i] = 6*v2*x 
		result[0,2,2,i] = 0 
		result[1,0,0,i] = 0 
		result[1,0,1,i] = 6*v3*x 
		result[1,0,2,i] = 6*v2*x 
		result[1,1,0,i] = 6*v3*x 
		result[1,1,1,i] = 0 
		result[1,1,2,i] = 6*v1*x 
		result[1,2,0,i] = 6*v2*x 
		result[1,2,1,i] = 6*v1*x 
		result[1,2,2,i] = 0 
		result[2,0,0,i] = 0 
		result[2,0,1,i] = 6*v2*x 
		result[2,0,2,i] = 0 
		result[2,1,0,i] = 6*v2*x 
		result[2,1,1,i] = 6*v1*x 
		result[2,1,2,i] = 0 
		result[2,2,0,i] = 0 
		result[2,2,1,i] = 0 
		result[2,2,2,i] = 0 
	if len(x_all) == 1:
		result = result[:,:,:,0]
	return result 



import numpy as np
from scipy import stats
from scipy.optimize import minimize

from . import evaluate_dmgs_equation as cp_dmgs
from . import utils as cp_utils
from . import genextreme_libs as cp_gev_b
from . import genextreme_derivs as cp_gev_c


def ppf(x, p = np.arange(0.1, 1.0, 0.1),
           ics =  [0, 0, 0],
           fdalpha = 0.01,
           means = False,
           waicscores = False,
           extramodels = False,
           pdf = False,
           customprior: float = 0,
           dmgs = True,
           debug = False):
    """
    Passed data from the Generalized Extreme Value Distribution, returns quantiles and other results based on a Calibrating Prior.
    The calibrating prior we use is given by π(μ,σ,ξ) ∝ 1/σ as given in Jewson et al. (2025).

    Parameters
    ----------
    x : array_like
        Input data array.
    p : array_like, optional
        Probabilities for quantile calculation (default is np.arange(0.1, 1.0, 0.1)).
    ics : list of float, optional
        Initial parameter estimates for optimization (default is [0, 0, 0]).
    fdalpha : float, optional
        Finite difference step for PDF estimation (default is 0.01).
    means : bool, optional
        Whether to compute means for extra models (default is False).
    waicscores : bool, optional
        Whether to compute WAIC scores (default is False).
    extramodels : bool, optional
        Whether to compute extra models (default is False).
    pdf : bool, optional
        Whether to compute PDFs (default is False).
    customprior : float, optional
        Custom prior value for shape parameter (default is 0).
    dmgs : bool, optional
        Whether to use the DMGS method (default is True).
    debug : bool, optional
        If True, print debug information (default is False).

    Returns
    -------
    dict
        Dictionary containing ML parameters, quantiles, PDFs, means, WAIC scores, and other results.
    """
    
    # 1 intro
    # optional parameters removed until rust, pwm are implemented
    nrust = 1000
    pwm = False
    rust = False

    x = cp_utils.to_array(x)
    p = cp_utils.to_array(p)
    
    # Input validation
    assert np.all(np.isfinite(x)) and not np.any(np.isnan(x)), "x must be finite and not NaN"
    assert np.all(np.isfinite(p)) and not np.any(np.isnan(p)), "p must be finite and not NaN"
    assert np.all(p > 0) and np.all(p < 1), "p must be between 0 and 1"
    assert len(ics) == 3, "ics must have length 3"
    assert fdalpha < 1, "fdalpha must be less than 1"
    
    alpha = 1-p
    nx = len(x)
    
    if pdf:
        dalpha = np.minimum(fdalpha * alpha, fdalpha * (1 - alpha))
        alpham = alpha - dalpha
        alphap = alpha + dalpha
    

    # 2 ml param estimate
    ics = cp_gev_b.gev_setics(x, ics)
    
    # Use minimize instead of optim (with negative log-likelihood)
    opt1 = minimize(lambda params: -cp_gev_b.gev_loglik(params, x), ics, method='BFGS')
    v1hat = opt1.x[0]
    v2hat = opt1.x[1]  
    v3hat = opt1.x[2]
    ml_params = [v1hat, v2hat, v3hat]
    
    if debug:
        print(f"    v1hat,v2hat,v3hat={v1hat},{v2hat},{v3hat}")
    
    pw_params = "pwm not selected"
    if pwm:
        pw_params = cp_gev_b.gev_pwm_params(x)
    
    if abs(v3hat) >= 1:
        revert2ml = True
    else:
        revert2ml = False
    
    # 3 aic
    ml_value = -opt1.fun
    maic = cp_utils.make_maic(ml_value, nparams=3)
    
    # 4 calc ml quantiles and densities (vectorized over alpha)
    ml_quantiles = stats.genextreme.ppf(1 - alpha,-v3hat, loc=v1hat, scale=v2hat)
    
    if v3hat < 0:
        ml_max = v1hat - v2hat / v3hat
    else:
        ml_max = np.inf
    
    fhat = stats.genextreme.pdf(ml_quantiles, -v3hat, loc=v1hat, scale=v2hat)
    
    pw_quantiles = "pwm not selected"
    if pwm:
        pw_quantiles = stats.genextreme.ppf(1 - alpha,-pw_params[2],loc=pw_params[0], scale=pw_params[1])
    
    # dmgs
    standard_errors = "dmgs not selected"
    rh_flat_quantiles = "dmgs not selected"
    cp_quantiles = "dmgs not selected"
    ru_quantiles = "dmgs not selected"
    custom_quantiles = "dmgs not selected"
    ml_pdf = "dmgs not selected"
    rh_flat_pdf = "dmgs not selected"
    cp_pdf = "dmgs not selected"
    waic1 = "dmgs not selected"
    waic2 = "dmgs not selected"
    ml_mean = "dmgs not selected"
    rh_mean = "dmgs not selected"
    rh_flat_mean = "dmgs not selected"
    cp_mean = "dmgs not selected"
    cp_method = "dmgs not selected"
    custom_mean = "dmgs not selected"
    
    # 5 alpha pdf stuff
    if dmgs and not revert2ml:
        if debug:
            print(f"  ml_quantiles={ml_quantiles}")
        
        if pdf:
            ml_quantilesm = stats.genextreme.ppf(1 - alpham, -v3hat, loc=v1hat, scale=v2hat)
            ml_quantilesp = stats.genextreme.ppf(1 - alphap, -v3hat, loc=v1hat, scale=v2hat)
            fhatm = stats.genextreme.pdf(ml_quantilesm, -v3hat,loc=v1hat, scale=v2hat)
            fhatp = stats.genextreme.pdf(ml_quantilesp, -v3hat,loc=v1hat, scale=v2hat)
        
        # 7 ldd
        if debug:
            print("  calculate ldd")
        ldd = cp_gev_c.gev_ldda(x, v1hat, v2hat, v3hat)
        lddi = np.linalg.inv(ldd)
        standard_errors = cp_utils.make_se(nx, lddi)
        
        # 8 lddd
        if debug:
            print("  calculate lddd")
        lddd = cp_gev_c.gev_lddda(x, v1hat, v2hat, v3hat)
        
        # 9 mu1
        mu1 = cp_gev_c.gev_mu1fa(alpha, v1hat, v2hat, v3hat)
        if pdf:
            mu1m = cp_gev_c.gev_mu1fa(alpham, v1hat, v2hat, v3hat)
            mu1p = cp_gev_c.gev_mu1fa(alphap, v1hat, v2hat, v3hat)
        
        # 10 mu2
        mu2 = cp_gev_c.gev_mu2fa(alpha, v1hat, v2hat, v3hat)
        if pdf:
            mu2m = cp_gev_c.gev_mu2fa(alpham, v1hat, v2hat, v3hat)
            mu2p = cp_gev_c.gev_mu2fa(alphap, v1hat, v2hat, v3hat)
        
        # 13 model 4: rh_flat with flat prior on shape
        lambdad_rh_flat = np.asarray([0, -1/v2hat, 0])
        dq = cp_dmgs.dmgs(lddi, lddd, mu1, lambdad_rh_flat, mu2, dim=3)
        rh_flat_quantiles = ml_quantiles + dq / (nx * fhat)
        
        if pdf:
            dqm = cp_dmgs.dmgs(lddi, lddd, mu1m, lambdad_rh_flat, mu2m, dim=3)
            dqp = cp_dmgs.dmgs(lddi, lddd, mu1p, lambdad_rh_flat, mu2p, dim=3)
            quantilesm = ml_quantilesm + dqm / (nx * fhatm)
            quantilesp = ml_quantilesp + dqp / (nx * fhatp)
            ml_pdf = fhat
            rh_flat_pdf = -(alphap - alpham) / (quantilesp - quantilesm)
        else:
            ml_pdf = fhat
            rh_flat_pdf = "pdf not selected"
        
        # 15 model 6: custom prior on shape parameter
        if extramodels or means:
            lambdad_custom = np.asarray([0, -1/v2hat, customprior])
            dq = cp_dmgs.dmgs(lddi, lddd, mu1, lambdad_custom, mu2, dim=3)
            custom_quantiles = ml_quantiles + dq / (nx * fhat)
        else:
            custom_quantiles = "extramodels not selected"
            lambdad_custom = 0
        
        # 16 means
        # check that lambdad_custom=0 is the correct result if neither extramodels nor means holds
        means_result = cp_gev_b.gev_means(means, ml_params, lddi, lddd,
                                lambdad_rh_flat, lambdad_custom, nx, dim=3)
        ml_mean = means_result['ml_mean']
        rh_flat_mean = means_result['rh_flat_mean']
        custom_mean = means_result['custom_mean']
        
        # 17 waicscores
        waic = cp_gev_b.gev_waic(waicscores, x, v1hat, v2hat, v3hat, lddi, lddd, lambdad_rh_flat)
        waic1 = waic['waic1']
        waic2 = waic['waic2']
        
        # 19 rust
        ru_quantiles = "rust not selected"
        if rust:
            rustsim = rvs(nrust, x, rust=True, mlcp=False)
            ru_quantiles = cp_gev_b.makeq(rustsim['ru_deviates'], p)
        
        # end of if(dmgs)
    else:
        rh_flat_quantiles = ml_quantiles
        ru_quantiles = ml_quantiles
        pw_quantiles = ml_quantiles
        custom_quantiles = ml_quantiles
        rh_flat_pdf = ml_pdf
        rh_flat_mean = ml_mean
        custom_mean = ml_mean
    
    return {
        'ml_params': ml_params,
        'pw_params': pw_params,
        'ml_value': ml_value,
        'standard_errors': standard_errors,
        'ml_quantiles': ml_quantiles,
        'ml_max': ml_max,
        'revert2ml': revert2ml,
        'cp_quantiles': rh_flat_quantiles,
        'ru_quantiles': ru_quantiles,
        'pw_quantiles': pw_quantiles,
        'custom_quantiles': custom_quantiles,
        'ml_pdf': ml_pdf,
        'cp_pdf': rh_flat_pdf,
        'maic': maic,
        'waic1': waic1,
        'waic2': waic2,
        'ml_mean': ml_mean,
        'cp_mean': rh_flat_mean,
        'custom_mean': custom_mean,
        'cp_method': cp_utils.crhpflat_dmgs_cpmethod()
    }


def rvs(n, x, ics  = [0, 0, 0], extramodels = False, mlcp = True):
    """
    Passed data from the Generalized Extreme Value Distribution, generate random samples from the same distribution with calibrating prior.

    Parameters
    ----------
    n : int
        Number of samples to generate.
    x : array_like
        Input data array.
    ics : list of float, optional
        Initial parameter estimates for optimization (default is [0, 0, 0]).
    extramodels : bool, optional
        Whether to compute extra models (default is False).
    mlcp : bool, optional
        Whether to use ML and CP deviates (default is True).
    debug : bool, optional
        If True, print debug information (default is False).

    Returns
    -------
    dict
        Dictionary containing ML parameters, ML deviates, CP deviates, RU deviates, and method info.
    """
    
    # optional parameter removed until rust is implemented
    rust = False

    x = cp_utils.to_array(x)
    
    assert np.isfinite(n) and not np.isnan(n), "n must be finite and not NaN"
    assert np.all(np.isfinite(x)) and not np.any(np.isnan(x)), "x must be finite and not NaN"
    assert len(ics) == 3, "ics must have length 3"
    
    ml_params = "mlcp not selected"
    ml_deviates = "mlcp not selected"
    cp_deviates = "mlcp not selected"
    ru_deviates = "rust not selected"
    
    if mlcp:
        q = ppf(x, np.random.uniform(0, 1, n), ics=ics, extramodels=extramodels)
        ml_params = q['ml_params']
        ml_deviates = q['ml_quantiles']
        ru_deviates = q['ru_quantiles']
        cp_deviates = q['cp_quantiles']
    
    if rust:
        th = tgev_cp(n, x)['theta_samples']
        ru_deviates = np.zeros(n)
        for i in range(n):
            ru_deviates[i] = stats.genextreme.rvs(-th[i, 2], loc=th[i, 0], scale=th[i, 1])
    
    return {
        'ml_params': ml_params,
        'ml_deviates': ml_deviates,
        'cp_deviates': cp_deviates,
        'ru_deviates': ru_deviates,
        'cp_method': cp_utils.crhpflat_dmgs_cpmethod()
    }


def pdf(x, y = None, ics = [0, 0, 0]):
    """
    Passed data from the Generalized Extreme Value Distribution, compute the density function for the GEV distribution with calibrating prior.

    Parameters
    ----------
    x : array_like
        Input data array.
    y : array_like, optional
        Points at which to evaluate the density (default is x).
    ics : list of float, optional
        Initial parameter estimates for optimization (default is [0, 0, 0]).
    extramodels : bool, optional
        Whether to compute extra models (default is False).
    debug : bool, optional
        If True, print debug information (default is False).

    Returns
    -------
    dict
        Dictionary containing ML parameters, ML PDF, RU PDF, and method info.
    """

    # optional parameters removed until rust is implemented
    rust = False
    nrust = 1000
    
    x = cp_utils.to_array(x)
    if y is None:
        y = x
    else:
        y = cp_utils.to_array(y)
    
    assert np.all(np.isfinite(x)) and not np.any(np.isnan(x)), "x must be finite and not NaN"
    assert np.all(np.isfinite(y)) and not np.any(np.isnan(y)), "y must be finite and not NaN"
    assert len(ics) == 3, "ics must have length 3"
    
    ics = cp_gev_b.gev_setics(x, ics)
    opt1 = minimize(lambda params: -cp_gev_b.gev_loglik(params, x), ics, method='BFGS')
    v1hat = opt1.x[0]
    v2hat = opt1.x[1]
    v3hat = opt1.x[2]
    
    if v3hat <= -1:
        revert2ml = True
    else:
        revert2ml = False
    
    dd = cp_gev_b.dgevsub(x=x, y=y, ics=ics)
    ru_pdf = "rust not selected"
    
    if rust and not revert2ml:
        th = tgev_cp(nrust, x)['theta_samples']
        ru_pdf = np.zeros(len(y))
        for ir in range(nrust):
            ru_pdf = ru_pdf + stats.genextreme.pdf(y, -th[ir, 2],loc=th[ir, 0], scale=th[ir, 1])
        ru_pdf = ru_pdf / nrust
    else:
        ru_pdf = dd['ml_pdf']
    
    return {
        'ml_params': dd['ml_params'],
        'ml_pdf': dd['ml_pdf'],
        'revert2ml': revert2ml,
        'cp_pdf': ru_pdf,
        'cp_method': cp_utils.nopdfcdfmsg()
    }


def cdf(x, y = None,
           ics = [0, 0, 0],
           extramodels = False,
           debug = False):
    """
    Passed data from the Generalized Extreme Value Distribution, compute the cumulative distribution function for the GEV distribution with calibrating prior.

    Parameters
    ----------
    x : array_like
        Input data array.
    y : array_like, optional
        Points at which to evaluate the CDF (default is x).
    ics : list of float, optional
        Initial parameter estimates for optimization (default is [0, 0, 0]).
    extramodels : bool, optional
        Whether to compute extra models (default is False).
    debug : bool, optional
        If True, print debug information (default is False).

    Returns
    -------
    dict
        Dictionary containing ML parameters, ML CDF, RU CDF, and method info.
    """

    # optional parameters removed until rust is implemented
    rust = False
    nrust = 1000
    
    x = cp_utils.to_array(x)
    if y is None:
        y = x
    y = cp_utils.to_array(y)
    
    assert np.all(np.isfinite(x)) and not np.any(np.isnan(x)), "x must be finite and not NaN"
    assert np.all(np.isfinite(y)) and not np.any(np.isnan(y)), "y must be finite and not NaN"
    assert len(ics) == 3, "ics must have length 3"
    
    ics = cp_gev_b.gev_setics(x, ics)
    opt1 = minimize(lambda params: -cp_gev_b.gev_loglik(params, x), ics, method='BFGS')
    v1hat = opt1.x[0]
    v2hat = opt1.x[1]
    v3hat = opt1.x[2]
    
    if v3hat <= -1:
        revert2ml = True
    else:
        revert2ml = False
    
    ml_params = [v1hat, v2hat, v3hat]
    dd = cp_gev_b.dgevsub(x=x, y=y, ics=ics)
    ru_cdf = "rust not selected"
    
    if rust and not revert2ml:
        th = tgev_cp(nrust, x)['theta_samples']
        ru_cdf = np.zeros(len(y))
        for ir in range(nrust):
            ru_cdf = ru_cdf + stats.genextreme.cdf(y, -th[ir, 2], loc=th[ir, 0], scale=th[ir, 1])
        ru_cdf = ru_cdf / nrust
    else:
        ru_cdf = dd['ml_cdf']
    
    return {
        'ml_params': dd['ml_params'],
        'ml_cdf': dd['ml_cdf'],
        'revert2ml': revert2ml,
        'cp_cdf': ru_cdf,
        'cp_method': cp_utils.nopdfcdfmsg()
    }


# not yet implemented
def tgev_cp(n, x, ics = [0, 0, 0]):
    """
    Not yet implemented: Theta sampling for the GEV distribution with calibrating prior.

    Parameters
    ----------
    n : int
        Number of theta samples to generate.
    x : array_like
        Input data array.
    ics : list of float, optional
        Initial parameter estimates for optimization (default is [0, 0, 0]).
    extramodels : bool, optional
        Whether to compute extra models (default is False).
    debug : bool, optional
        If True, print debug information (default is False).

    Returns
    -------
    dict
        Dictionary containing theta samples.
    """
        
    raise Exception('tgev_cp (and rust) is not yet implemented in fitdistcp; please use the dmgs method.')
    
    x = cp_utils.to_array(x)
    
    assert np.all(np.isfinite(x)) and not np.any(np.isnan(x)), "x must be finite and not NaN"
    assert len(ics) == 3, "ics must have length 3"
    
    ics = cp_gev_b.gev_setics(x, ics)
    t = ru(cp_gev_b.gev_logf, x=x, n=n, d=3, init=ics)
    
    return {'theta_samples': t['sim_vals']}


import numpy as np

def exp_fd(x, v1): 
	#Automatically generated code.
	dim=1 
	if not (isinstance(x, list) or isinstance(x, np.ndarray)):
		x_all=[x]
	else:
		x_all=x
	result = np.zeros((dim, len(x_all)))
	for i in range(len(x_all)):
		x = x_all[i]
		result[0,i] = -v1*x*np.exp(-v1*x) + np.exp(-v1*x) 
	if len(x_all) == 1:
		result = result[:,0]
	return result 

def exp_fdd(x, v1): 
	#Automatically generated code.
	dim=1 
	if not (isinstance(x, list) or isinstance(x, np.ndarray)):
		x_all=[x]
	else:
		x_all=x
	result = np.zeros((dim,dim, len(x_all)))
	for i in range(len(x_all)):
		x = x_all[i]
		result[0,0,i] = x*(v1*x - 2)*np.exp(-v1*x)
	if len(x_all) == 1:
		result = result[:,:,0]
	return result 

def exp_pd(x, v1): 
	#Automatically generated code.
	dim=1 
	if not (isinstance(x, list) or isinstance(x, np.ndarray)):
		x_all=[x]
	else:
		x_all=x
	result = np.zeros((dim, len(x_all)))
	for i in range(len(x_all)):
		x = x_all[i]
		result[0,i] = x*np.exp(-v1*x) 
	if len(x_all) == 1:
		result = result[:,0]
	return result 

def exp_pdd(x, v1): 
	#Automatically generated code.
	dim=1 
	if not (isinstance(x, list) or isinstance(x, np.ndarray)):
		x_all=[x]
	else:
		x_all=x
	result = np.zeros((dim,dim, len(x_all)))
	for i in range(len(x_all)):
		x = x_all[i]
		result[0,0,i] = -x**2*np.exp(-v1*x)
	if len(x_all) == 1:
		result = result[:,:,0]
	return result 

def exp_logfdd(x, v1): 
	#Automatically generated code.
	dim=1 
	if not (isinstance(x, list) or isinstance(x, np.ndarray)):
		x_all=[x]
	else:
		x_all=x
	result = np.zeros((dim,dim, len(x_all)))
	for i in range(len(x_all)):
		x = x_all[i]
		result[0,0,i] = -1/v1**2
	if len(x_all) == 1:
		result = result[:,:,0]
	return result 

def exp_logfddd(x, v1): 
	#Automatically generated code.
	dim=1 
	if not (isinstance(x, list) or isinstance(x, np.ndarray)):
		x_all=[x]
	else:
		x_all=x
	result = np.zeros((dim, dim, dim, len(x_all)))
	for i in range(len(x_all)):
		x = x_all[i]
		result[0,0,0,i] = 2/v1**3 
	if len(x_all) == 1:
		result = result[:,:,:,0]
	return result 


# The first derivative of the density
def exp_f1fa(x,v1):
	nx=len(x)
	f1=np.zeros((1,nx))
	f1[0,:]=exp_fd(x,v1)
	return f1

# The second derivative of the density
def exp_f2fa(x,v1):
	nx=len(x)
	f2 = np.zeros((1, 1, nx))
	f2[0,0,:] = exp_fdd(x,v1)
	return f2

# The first derivative of the cdf
def exp_p1fa(x,v1):
	nx=len(x)
	p1 = np.zeros((1, nx))
	p1[0,:]=exp_pd(x,v1)
	return p1

# The second derivative of the cdf
def exp_p2fa(x,v1):
	nx=len(x)
	p2=np.zeros((1,1,nx))
	p2[0,0,:]=exp_pdd(x,v1)
	return p2

# The second derivative of the normalized log-likelihood
def exp_ldda(x,v1):
	nx=len(x)
	ldd=np.zeros((1,1))
	ldd[0,0]=np.sum(exp_logfdd(x,v1), axis=-1)/nx
	return ldd

# The third derivative of the normalized log-likelihood
def exp_lddda(x,v1):
	nx=len(x)
	lddd = np.zeros((1,1,1))
	lddd[0,0,0]=np.sum(exp_logfddd(x,v1), axis=-1)/nx
	return lddd



import numpy as np
import scipy.stats as stats

from . import norm_libs
from . import utils as cp_utils


def ppf(x, p=np.arange(0.1, 1.0, 0.1), means=False, waicscores=False, 
             logscores=False, rust=False, nrust=100000, unbiasedv=False, debug=False):
    """
    Normal Distribution Predictions Based on a Calibrating Prior
    
    Parameters
    ----------
    x : array-like
        Training data
    p : array-like, default np.arange(0.1, 1.0, 0.1)
        Probability levels for quantiles
    means : bool, default False
        Whether to calculate means
    waicscores : bool, default False
        Whether to calculate WAIC scores
    logscores : bool, default False
        Whether to calculate log scores
    rust : bool, default False
        Whether to use RUST method
    nrust : int, default 100000
        Number of RUST samples
    unbiasedv : bool, default False
        Whether to use unbiased variance
    debug : bool, default False
        Debug flag
        
    Returns
    -------
    dict
        Dictionary containing quantiles and other statistics
    """
    # 1 intro
    x = cp_utils.to_array(x)
    p = cp_utils.to_array(p)
    assert np.all(np.isfinite(x)) and not np.any(np.isnan(x))
    assert np.all(np.isfinite(p)) and not np.any(np.isnan(p))
    assert np.all(p > 0) and np.all(p < 1)
    
    alpha = 1 - p
    nx = len(x)
    nalpha = len(alpha)
    
    # 2 ml param estimate
    ml_params = norm_libs.norm_ml_params(x)
    v1hat = ml_params[0]
    v2hat = ml_params[1]
    uv_params = "unbiasedv not selected"
    if unbiasedv:
        uv_params = norm_libs.norm_unbiasedv_params(x)
    
    # 3 aic
    ml_value = np.sum(stats.norm.logpdf(x, loc=v1hat, scale=v2hat))
    maic = cp_utils.make_maic(ml_value, nparams=2)
    
    # 4 ml quantiles (vectorized over alpha)
    ml_quantiles = stats.norm.ppf(1 - alpha, loc=v1hat, scale=v2hat)
    uv_quantiles = "unbiasedv not selected"
    if unbiasedv:
        uv_quantiles = stats.norm.ppf(1 - alpha, loc=uv_params[0], scale=uv_params[1])
    
    # 5 rhp quantiles (vectorized over alpha)
    mu = v1hat
    
    # first, convert sigma from maxlik to unbiased
    sgu = v2hat * np.sqrt(nx / (nx - 1))
    # then, convert sigma to predictive sigma
    sg = sgu * np.sqrt((nx + 1) / nx)
    
    temp = stats.t.ppf(1 - alpha, df=nx - 1)
    rh_quantiles = mu + temp * sg
    
    ldd = "only relevant for DMGS models, not analytic models"
    lddi = "only relevant for DMGS models, not analytic models"
    expinfmat = np.zeros((2, 2))
    expinfmat[0, 0] = nx / (v2hat * v2hat)
    expinfmat[1, 1] = 2 * nx / (v2hat * v2hat)
    expinfmati = np.linalg.inv(expinfmat)
    standard_errors = np.zeros(2)
    standard_errors[0] = np.sqrt(expinfmati[0, 0])
    standard_errors[1] = np.sqrt(expinfmati[1, 1])
    
    # test of gg code (for future implementation of mpd theory, as a test of the mpd code)
    # norm_gg(nx,v1hat,v2hat)
    
    # 6 means (might as well always calculate)
    ml_mean = v1hat
    rh_mean = v1hat
    
    # 7 waicscores
    waic = norm_libs.norm_waic(waicscores, x, v1hat, v2hat)
    waic1 = waic['waic1']
    waic2 = waic['waic2']
    
    # 8 logscores
    logscores_result = norm_libs.norm_logscores(logscores, x)
    ml_oos_logscore = logscores_result['ml_oos_logscore']
    rh_oos_logscore = logscores_result['rh_oos_logscore']
    
    # 9 rust
    ru_quantiles = "rust not selected"
    if rust:
        rustsim = rvs(nrust, x, rust=True, mlcp=False)
        ru_quantiles = cp_utils.makeq(rustsim['ru_deviates'], p)
    
    return {
        'ml_params': ml_params,
        'ml_value': ml_value,
        'uv_params': uv_params,
        # 'ldd': ldd,
        # 'lddi': lddi,
        # 'expinfmat': expinfmat,
        # 'expinfmati': expinfmati,
        'standard_errors': standard_errors,
        'ml_quantiles': ml_quantiles,
        'cp_quantiles': rh_quantiles,
        'ru_quantiles': ru_quantiles,
        'uv_quantiles': uv_quantiles,
        'maic': maic,
        'waic1': waic1,
        'waic2': waic2,
        'ml_oos_logscore': ml_oos_logscore,
        'cp_oos_logscore': rh_oos_logscore,
        'ml_mean': ml_mean,
        'cp_mean': rh_mean,
        'cp_method': cp_utils.analytic_cpmethod()
    }

def rvs(n, x, rust=False, mlcp=True, debug=False):
    """
    Random number generation for normal distribution with calibrating prior
    
    Parameters
    ----------
    n : int
        Number of samples
    x : array-like
        Training data
    rust : bool, default False
        Whether to use RUST method
    mlcp : bool, default True
        Whether to use ML/CP method
    debug : bool, default False
        Debug flag
        
    Returns
    -------
    dict
        Dictionary containing random deviates
    """
    # stopifnot(is.finite(n),!is.na(n),is.finite(x),!is.na(x))
    x = cp_utils.to_array(x)
    assert np.all(np.isfinite(x)) and not np.any(np.isnan(x))
    
    ml_params = "mlcp not selected"
    ml_deviates = "mlcp not selected"
    cp_deviates = "mlcp not selected"
    ru_deviates = "rust not selected"
    
    if mlcp:
        q = ppf(x, np.random.uniform(size=n))
        ml_params = q['ml_params']
        ml_deviates = q['ml_quantiles']
        cp_deviates = q['cp_quantiles']
    
    if rust:
        th = tnorm_cp(n, x)['theta_samples']
        ru_deviates = np.zeros(n)
        for i in range(n):
            ru_deviates[i] = np.random.normal(loc=th[i, 0], scale=th[i, 1])
    
    op = {
        'ml_params': ml_params,
        'ml_deviates': ml_deviates,
        'cp_deviates': cp_deviates,
        'ru_deviates': ru_deviates,
        'cp_method': cp_utils.analytic_cpmethod()
    }
    return op

def pdf(x, y=None, rust=False, nrust=1000, debug=False):
    """
    Density function for normal distribution with calibrating prior
    
    Parameters
    ----------
    x : array-like
        Training data
    y : array-like, optional
        Test points (default: same as x)
    rust : bool, default False
        Whether to use RUST method
    nrust : int, default 1000
        Number of RUST samples
    debug : bool, default False
        Debug flag
        
    Returns
    -------
    dict
        Dictionary containing density values
    """
    x = cp_utils.to_array(x)
    if y is None:
        y = x
    y = cp_utils.to_array(y)
    
    assert np.all(np.isfinite(x)) and not np.any(np.isnan(x))
    assert np.all(np.isfinite(y)) and not np.any(np.isnan(y))
    
    dd = norm_libs.dnormsub(x=x, y=y)
    ru_pdf = "rust not selected"
    if rust:
        th = tnorm_cp(nrust, x)['theta_samples']
        ru_pdf = np.zeros(len(y))
        for ir in range(nrust):
            ru_pdf = ru_pdf + stats.norm.pdf(y, loc=th[ir, 0], scale=th[ir, 1])
        ru_pdf = ru_pdf / nrust
    
    op = {
        'ml_params': dd['ml_params'],
        'ml_pdf': dd['ml_pdf'],
        'cp_pdf': dd['rh_pdf'],
        'ru_pdf': ru_pdf,
        'cp_method': cp_utils.analytic_cpmethod()
    }
    return op

def cdf(x, y=None, rust=False, nrust=1000, debug=False):
    """
    Cumulative distribution function for normal distribution with calibrating prior
    
    Parameters
    ----------
    x : array-like
        Training data
    y : array-like, optional
        Test points (default: same as x)
    rust : bool, default False
        Whether to use RUST method
    nrust : int, default 1000
        Number of RUST samples
    debug : bool, default False
        Debug flag
        
    Returns
    -------
    dict
        Dictionary containing CDF values
    """
    x = cp_utils.to_array(x)
    if y is None:
        y = x
    y = cp_utils.to_array(y)
    
    assert np.all(np.isfinite(x)) and not np.any(np.isnan(x))
    assert np.all(np.isfinite(y)) and not np.any(np.isnan(y))
    
    dd = norm_libs.dnormsub(x=x, y=y)
    ru_cdf = "rust not selected"
    if rust:
        th = tnorm_cp(nrust, x)['theta_samples']
        ru_cdf = np.zeros(len(y))
        for ir in range(nrust):
            ru_cdf = ru_cdf + stats.norm.cdf(y, loc=th[ir, 0], scale=th[ir, 1])
        ru_cdf = ru_cdf / nrust
    
    op = {
        'ml_params': dd['ml_params'],
        'ml_cdf': dd['ml_cdf'],
        'cp_cdf': dd['rh_cdf'],
        'ru_cdf': ru_cdf,
        'cp_method': cp_utils.analytic_cpmethod()
    }
    return op

def tnorm_cp(n, x, debug=False):
    """
    Not yet implemented: Theta sampling for normal distribution with calibrating prior
    
    Parameters
    ----------
    n : int
        Number of samples
    x : array-like
        Training data
    debug : bool, default False
        Debug flag
        
    Returns
    -------
    dict
        Dictionary containing theta samples
    """
    # stopifnot(is.finite(n),!is.na(n),is.finite(x),!is.na(x))
    x = cp_utils.to_array(x)
    assert np.all(np.isfinite(x)) and not np.any(np.isnan(x))
    
    t = cp_utils.ru(norm_libs.norm_logf, x=x, n=n, d=2, init=[np.mean(x), np.std(x)])
    
    return {'theta_samples': t['sim_vals']}
